"""
Functionality to handle different kinds of reaction SMILES with the same functions.

In a separate file than miscellaneous.py or conversion.py in order to avoid
cyclic dependencies.
"""

from enum import auto

from rxn.utilities.types import RxnEnum

from .extended_reaction_smiles import (
    parse_extended_reaction_smiles,
    to_extended_reaction_smiles,
)
from .reaction_equation import ReactionEquation


class ReactionFormat(RxnEnum):
    """
    Existing reaction SMILES formats.

    Attributes:
        EXTENDED: extended reaction SMILES with fragment info,
            f.i. ``|f:0.2,3.4.5|``.
        STANDARD: standard reaction SMILES.
        STANDARD_WITH_TILDE: standard reaction SMILES, where fragments are
            indicated with tilde symbols, "~".
    """

    EXTENDED = auto()
    STANDARD = auto()
    STANDARD_WITH_TILDE = auto()


def determine_format(reaction_smiles: str) -> ReactionFormat:
    """
    Determine the format of a reaction SMILES.
    """
    if " |" in reaction_smiles:
        return ReactionFormat.EXTENDED

    if "~" in reaction_smiles:
        return ReactionFormat.STANDARD_WITH_TILDE

    return ReactionFormat.STANDARD


def parse_any_reaction_smiles(smiles: str) -> ReactionEquation:
    """
    Parse a reaction SMILES in any format (will be determined automatically).
    """
    return parse_reaction_smiles(smiles, reaction_format=determine_format(smiles))


def parse_reaction_smiles(
    smiles: str, reaction_format: ReactionFormat
) -> ReactionEquation:
    """
    Parse the reaction SMILES in a given format.
    """
    if reaction_format is ReactionFormat.EXTENDED:
        return parse_extended_reaction_smiles(smiles, remove_atom_maps=False)

    if reaction_format is ReactionFormat.STANDARD:
        return ReactionEquation.from_string(smiles)

    if reaction_format is ReactionFormat.STANDARD_WITH_TILDE:
        return ReactionEquation.from_string(smiles, fragment_bond="~")

    raise ValueError(f"Unsupported reaction format: {reaction_format}")


def to_reaction_smiles(
    reaction_equation: ReactionEquation, reaction_format: ReactionFormat
) -> str:
    """
    Convert a reaction equation into a reaction SMILES of the specified format.
    """
    if reaction_format is ReactionFormat.EXTENDED:
        return to_extended_reaction_smiles(reaction_equation)

    if reaction_format is ReactionFormat.STANDARD:
        return reaction_equation.to_string()

    if reaction_format is ReactionFormat.STANDARD_WITH_TILDE:
        return reaction_equation.to_string(fragment_bond="~")

    raise ValueError(f"Unsupported reaction format: {reaction_format}")



import logging
import re
import shutil
from typing import List, Optional

from rxn.utilities.files import (
    PathLike,
    dump_list_to_file,
    iterate_lines_from_file,
    raise_if_paths_are_identical,
)

from .exceptions import UnclearWhetherTokenized

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


SMILES_TOKENIZER_PATTERN = r"(\%\([0-9]{3}\)|\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\||\(|\)|\.|=|#|-|\+|\\|\/|:|~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"
SMILES_REGEX = re.compile(SMILES_TOKENIZER_PATTERN)


class TokenizationError(ValueError):
    """Exception raised in RDKit."""

    def __init__(self, title: str, detail: str):
        """
        Initialize TokenizationError.

        Args:
            title: title of the error.
            detail: decscription of the error.
        """
        self.type = "TokenizationError"
        self.title = title
        self.detail = detail


def to_tokens(smiles: str) -> List[str]:
    """
    Tokenize a SMILES molecule or reaction into a list of tokens.

    Args:
        smiles: SMILES string to tokenize.

    Raises:
        TokenizationError: in case of mismatch between the SMILES and the joined tokens.

    Returns:
        List of tokens (give back the original SMILES string if appended).
    """
    tokens = [token for token in SMILES_REGEX.findall(smiles)]

    if smiles != "".join(tokens):
        raise TokenizationError(
            "SmilesJoinedTokensMismatch",
            f'SMILES="{smiles}" != joined_tokens="{"".join(tokens)}"',
        )

    return tokens


def tokenize_smiles(smiles: str, fallback_value: Optional[str] = None) -> str:
    """
    Tokenize a SMILES molecule or reaction, and join the tokens with spaces.

    Args:
        smiles: SMILES string to tokenize, for instance 'CC(CO)=N>>CC(C=O)N'.
        fallback_value: what value to returns when the tokenization is unsuccessful.
            Default: no fallback, will propagate the TokenizationError exception.

    Returns:
        SMILES string after tokenization, for instance 'C C ( C O ) = N >> C C ( C = O ) N'.
    """
    try:
        tokens = to_tokens(smiles)
        return " ".join(tokens)
    except TokenizationError:
        if fallback_value is not None:
            logger.debug(f'Error when tokenizing "{smiles}"')
            return fallback_value
        raise


def detokenize_smiles(tokenized_smiles: str) -> str:
    """
    Detokenize a tokenized SMILES string (that contains spaces between the characters).

    Args:
        tokenized_smiles: tokenized SMILES, for instance 'C C ( C O ) = N >> C C ( C = O ) N'

    Returns:
        SMILES after detokenization, for instance 'CC(CO)=N>>CC(C=O)N'
    """
    return tokenized_smiles.replace(" ", "")


def string_is_tokenized(smiles_line: str) -> bool:
    """
    Whether a string is a tokenized SMILES or not.

    Args:
        smiles_line: string to inspect

    Raises:
        ValueError: if not possible to determine whether tokenized or not
        TokenizationError: propagated directly from tokenize_smiles()
    """
    detokenized = detokenize_smiles(smiles_line)
    tokens = to_tokens(detokenized)
    if len(tokens) < 2:
        raise UnclearWhetherTokenized(smiles_line)
    return " ".join(tokens) == smiles_line


def tokenize_file(
    input_file: PathLike, output_file: PathLike, fallback_value: str = ""
) -> None:
    """
    Tokenize a file containing SMILES strings.

    Args:
        input_file: file to tokenize.
        output_file: where to save the tokenized file.
        fallback_value: placeholder for strings that cannot be tokenized.
    """
    raise_if_paths_are_identical(input_file, output_file)
    logger.info(f'Tokenizing "{input_file}" -> "{output_file}".')

    tokenized = (
        tokenize_smiles(line, fallback_value)
        for line in iterate_lines_from_file(input_file)
    )

    dump_list_to_file(tokenized, output_file)


def detokenize_file(
    input_file: PathLike,
    output_file: PathLike,
) -> None:
    raise_if_paths_are_identical(input_file, output_file)
    logger.info(f'Detokenizing "{input_file}" -> "{output_file}".')

    detokenized = (
        detokenize_smiles(line) for line in iterate_lines_from_file(input_file)
    )
    dump_list_to_file(detokenized, output_file)


def ensure_tokenized_file(
    file: PathLike, postfix: str = ".tokenized", fallback_value: str = ""
) -> str:
    """
    Ensure that a file is tokenized: do nothing if the file is already tokenized, create
    a tokenized copy otherwise.

    Args:
        file: path to the file that we want to ensure is tokenized.
        postfix: postfix to add to the tokenized copy (if applicable).
        fallback_value: placeholder for strings that cannot be tokenized (if applicable).

    Returns:
        The path to the tokenized file (original path, or path to new file).
    """
    if file_is_tokenized(file):
        return str(file)

    tokenized_copy = str(file) + postfix
    tokenize_file(file, tokenized_copy, fallback_value=fallback_value)
    return tokenized_copy


def file_is_tokenized(filepath: PathLike) -> bool:
    """
    Whether a file contains tokenized SMILES or not.

    By default, this looks at the first non-empty line of the file only!

    Raises:
        TokenizationError: propagated from tokenize_smiles()
        RuntimeError: for empty files or files with empty lines only.

    Args:
        filepath: path to the file.
    """
    # Iterative formulation in case the first line(s) of the file don't make it
    # clear whether tokenized or not.
    for line in iterate_lines_from_file(filepath):
        try:
            return string_is_tokenized(line)
        except UnclearWhetherTokenized:
            continue
    raise RuntimeError(
        f'Could not determine whether "{filepath}" is tokenized: empty lines only.'
    )


def copy_as_detokenized(src: PathLike, dest: PathLike) -> None:
    """
    Copy a source file to a destination, while making sure that it is not tokenized.
    """
    if file_is_tokenized(src):
        logger.info(f'Copying and detokenizing "{src}" -> "{dest}".')
        detokenize_file(src, dest)
    else:
        logger.info(f'Copying "{src}" -> "{dest}".')
        shutil.copy(src, dest)



import pytest

from rxn.chemutils.reaction_equation import ReactionEquation
from rxn.chemutils.reaction_smiles import (
    ReactionFormat,
    determine_format,
    parse_any_reaction_smiles,
    parse_reaction_smiles,
    to_reaction_smiles,
)


def test_determine_format() -> None:
    reaction_smiles_with_other_extended_information = (
        "[CH3:1][C:2](=[O:3])[NH:4][C@H:5]([C:6](=[O:7])[OH:8])[C:9]([CH3:10])"
        "([CH3:11])[SH:12].[O:13]=[N:14]O[Na]>O.Cl.CO>[CH3:1][C:2](=[O:3])[NH:4]"
        "[C@H:5]([C:6](=[O:7])[OH:8])[C:9]([CH3:10])([CH3:11])[S:12][N:14]=[O:13]"
        " |c:5,7,H:3.2,&1:4,24,f:3.4|"
    )

    reaction_smiles_with_other_extended_information_2 = (
        "[CH3:1][C:2](=[O:3])[NH:4][C@H:5]([C:6](=[O:7])[OH:8])[C:9]([CH3:10])"
        "([CH3:11])[SH:12].[O:13]=[N:14]O[Na]>O.Cl.CO>[CH3:1][C:2](=[O:3])[NH:4]"
        "[C@H:5]([C:6](=[O:7])[OH:8])[C:9]([CH3:10])([CH3:11])[S:12][N:14]=[O:13]"
        " |c:5,7,H:3.2,&1:4,24|"
    )

    reaction_smiles_with_other_extended_information_3 = (
        "[CH3:1][C:2](=[O:3])[NH:4][C@H:5]([C:6](=[O:7])[OH:8])[C:9]([CH3:10])"
        "([CH3:11])[SH:12].[O:13]=[N:14]O[Na]>O.Cl.CO>[CH3:1][C:2](=[O:3])[NH:4]"
        "[C@H:5]([C:6](=[O:7])[OH:8])[C:9]([CH3:10])([CH3:11])[S:12][N:14]=[O:13]"
        " |&1:4,24|"
    )

    assert determine_format("CC.O>>CCO") is ReactionFormat.STANDARD
    assert (
        determine_format("CC.O.[Na+]~[Cl-]>>CCO") is ReactionFormat.STANDARD_WITH_TILDE
    )
    assert determine_format("CC.O.[Na+].[Cl-]>>CCO |f:2.3|") is ReactionFormat.EXTENDED
    assert (
        determine_format(reaction_smiles_with_other_extended_information)
        is ReactionFormat.EXTENDED
    )
    assert (
        determine_format(reaction_smiles_with_other_extended_information_2)
        is ReactionFormat.EXTENDED
    )
    assert (
        determine_format(reaction_smiles_with_other_extended_information_3)
        is ReactionFormat.EXTENDED
    )


def test_parse_any_reaction_smiles() -> None:
    assert parse_any_reaction_smiles("CC.O>>CCO") == ReactionEquation(
        ["CC", "O"], [], ["CCO"]
    )
    assert parse_any_reaction_smiles("CC.O.[Na+]~[Cl-]>>CCO") == ReactionEquation(
        ["CC", "O", "[Na+].[Cl-]"], [], ["CCO"]
    )
    assert parse_any_reaction_smiles(
        "CC.O.[Na+].[Cl-]>>CCO |f:2.3|"
    ) == ReactionEquation(["CC", "O", "[Na+].[Cl-]"], [], ["CCO"])


def test_parse_any_reaction_smiles_with_mapping() -> None:
    # Molecules with mapping
    reactant_1 = "[CH3:1][CH3:2]"
    reactant_2 = "[OH2:3]"
    product = "[CH3:1][CH2:2][OH:3]"

    # In the extracted equation, mapping is still present.
    expected = ReactionEquation([reactant_1, reactant_2, "[Na+].[Cl-]"], [], [product])

    assert (
        parse_any_reaction_smiles(f"{reactant_1}.{reactant_2}.[Na+]~[Cl-]>>{product}")
        == expected
    )
    assert (
        parse_any_reaction_smiles(
            f"{reactant_1}.{reactant_2}.[Na+].[Cl-]>>{product} |f:2.3|"
        )
        == expected
    )


def test_parse_reaction_smiles() -> None:
    assert parse_reaction_smiles(
        "CC.O>>CCO", ReactionFormat.STANDARD
    ) == ReactionEquation(["CC", "O"], [], ["CCO"])
    assert parse_reaction_smiles(
        "CC.O.[Na+]~[Cl-]>>CCO", ReactionFormat.STANDARD_WITH_TILDE
    ) == ReactionEquation(["CC", "O", "[Na+].[Cl-]"], [], ["CCO"])
    assert parse_reaction_smiles(
        "CC.O.[Na+].[Cl-]>>CCO |f:2.3|", ReactionFormat.EXTENDED
    ) == ReactionEquation(["CC", "O", "[Na+].[Cl-]"], [], ["CCO"])

    with pytest.raises(ValueError):
        _ = parse_reaction_smiles("CC.O>>CCO", "standard")  # type: ignore


def test_to_reaction_smiles() -> None:
    reaction_equation = ReactionEquation(["CC", "O", "[Na+].[Cl-]"], [], ["CCO"])
    assert (
        to_reaction_smiles(reaction_equation, ReactionFormat.STANDARD)
        == "CC.O.[Na+].[Cl-]>>CCO"
    )
    assert (
        to_reaction_smiles(reaction_equation, ReactionFormat.STANDARD_WITH_TILDE)
        == "CC.O.[Na+]~[Cl-]>>CCO"
    )
    assert (
        to_reaction_smiles(reaction_equation, ReactionFormat.EXTENDED)
        == "CC.O.[Na+].[Cl-]>>CCO |f:2.3|"
    )

    with pytest.raises(ValueError):
        _ = to_reaction_smiles(reaction_equation, "standard")  # type: ignore



import re
from functools import partial
from typing import (
    Callable,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
)

import attr
from rxn.utilities.containers import remove_duplicates

from .conversion import canonicalize_smiles, cleanup_smiles
from .exceptions import InvalidReactionSmiles
from .multicomponent_smiles import (
    list_to_multicomponent_smiles,
    multicomponent_smiles_to_list,
)

T = TypeVar("T", bound="ReactionEquation")


@attr.s(auto_attribs=True, init=False)
class ReactionEquation:
    """
    Defines a reaction equation, as given by the molecules involved in a reaction.

    Attributes:
        reactants: SMILES strings for compounds on the left of the reaction arrow.
        agents: SMILES strings for compounds above the reaction arrow. Are
            sometimes merged with the reactants.
        products: SMILES strings for compounds on the right of the reaction arrow.
    """

    reactants: List[str]
    agents: List[str]
    products: List[str]

    def __init__(
        self, reactants: Iterable[str], agents: Iterable[str], products: Iterable[str]
    ):
        """
        Overwrite init function in order to enable instantiation from any iterator and
        to force copying the lists.
        """
        self.__attrs_init__(list(reactants), list(agents), list(products))

    def __iter__(self) -> Iterator[List[str]]:
        """Helper function to simplify functionality acting on all three
        compound groups"""
        return (i for i in (self.reactants, self.agents, self.products))

    def iter_all_smiles(self) -> Generator[str, None, None]:
        """Helper function to iterate over all the SMILES in the reaction equation"""
        return (molecule for group in self for molecule in group)

    def to_string(self, fragment_bond: Optional[str] = None) -> str:
        """
        Convert a ReactionEquation to an "rxn" reaction SMILES.
        """

        smiles_groups = (
            list_to_multicomponent_smiles(group, fragment_bond) for group in self
        )
        return ">".join(smiles_groups)

    @classmethod
    def from_string(
        cls: Type[T], reaction_string: str, fragment_bond: Optional[str] = None
    ) -> T:
        """
        Convert a ReactionEquation from an "rxn" reaction SMILES.
        """

        # We split at the ">" characters, only if they are not preceded by a "-",
        # which would indicate a dative bond.
        groups = [
            multicomponent_smiles_to_list(smiles_group, fragment_bond=fragment_bond)
            for smiles_group in re.split(r"(?<!-)>", reaction_string)
        ]

        try:
            return cls(*groups)
        except TypeError as e:
            raise InvalidReactionSmiles(reaction_string) from e


def merge_reactants_and_agents(reaction: ReactionEquation) -> ReactionEquation:
    return ReactionEquation(
        reactants=reaction.reactants + reaction.agents,
        agents=[],
        products=reaction.products,
    )


def sort_compounds(reaction: ReactionEquation) -> ReactionEquation:
    """
    Reorder the compounds of each group in alphabetic order.
    """
    return apply_to_compound_groups(reaction, sorted)


def apply_to_compounds(
    reaction: ReactionEquation, fn: Callable[[str], str]
) -> ReactionEquation:
    """
    Apply a function to the individual compounds in a ReactionEquation.

    Args:
        reaction: reaction equation to apply the function to.
        fn: function to apply.

    Returns:
        New ReactionEquation instance after application of the function to the compounds.
    """
    updated_compound_groups = (
        [fn(compound) for compound in compound_group] for compound_group in reaction
    )
    return ReactionEquation(*updated_compound_groups)


def apply_to_compound_groups(
    reaction: ReactionEquation, fn: Callable[[List[str]], List[str]]
) -> ReactionEquation:
    """
    Apply a function to the groups of compounds in a ReactionEquation.

    Args:
        reaction: reaction equation to apply the function to.
        fn: function to apply.

    Returns:
        New ReactionEquation instance after application of the function to the
        compound groups.
    """
    updated_compound_groups = (fn(compound_group) for compound_group in reaction)
    return ReactionEquation(*updated_compound_groups)


def canonicalize_compounds(
    reaction: ReactionEquation, check_valence: bool = True
) -> ReactionEquation:
    """
    Canonicalize the molecules of a ReactionEquation.
    """
    canonicalize_fn = partial(canonicalize_smiles, check_valence=check_valence)
    return apply_to_compounds(reaction, canonicalize_fn)


def remove_duplicate_compounds(reaction: ReactionEquation) -> ReactionEquation:
    """
    Remove compounds that are duplicated in the same category
    """
    return apply_to_compound_groups(reaction, remove_duplicates)


def cleanup_compounds(reaction: ReactionEquation) -> ReactionEquation:
    """
    Basic cleanup of the compounds.
    """
    return apply_to_compounds(reaction, cleanup_smiles)


def rxn_standardization(reaction: ReactionEquation) -> ReactionEquation:
    """
    Apply the standard RXN postprocessing of reaction equations.

    Consists in the following
    1. merge reactants and agents
    2. canonicalize all the SMILES
    3. sort the compounds in each group
    4. remove the duplicates
    """
    return remove_duplicate_compounds(
        sort_compounds(canonicalize_compounds(merge_reactants_and_agents(reaction)))
    )


def remove_precursors_from_products(reaction: ReactionEquation) -> ReactionEquation:
    """
    Remove compounds in products that are also present in the reactants or reagents.
    """
    precursors = reaction.reactants + reaction.agents
    products_without_precursors = [
        product for product in reaction.products if product not in precursors
    ]
    return ReactionEquation(
        reactants=reaction.reactants,
        agents=reaction.agents,
        products=products_without_precursors,
    )


def has_repeated_molecules(reaction_equation: ReactionEquation) -> bool:
    all_molecules = list(reaction_equation.iter_all_smiles())
    return len(set(all_molecules)) < len(all_molecules)



import operator
import re
from functools import reduce
from typing import List, Optional, Tuple, Union

from rdkit import Chem, RDLogger
from rdkit.Chem import (
    AssignStereochemistry,
    MolFromInchi,
    RemoveHs,
    SanitizeFlags,
    SanitizeMol,
)
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdmolfiles import (
    MolFromMolBlock,
    MolFromSmiles,
    MolToMolBlock,
    MolToSmiles,
)

from .exceptions import InvalidInchi, InvalidMdl, InvalidSmiles, SanitizationError

RDLogger.logger().setLevel(RDLogger.CRITICAL)  # type: ignore[no-untyped-call]


def smiles_to_mol(
    smiles: str, sanitize: bool = True, find_radicals: bool = True
) -> Mol:
    """
    Convert a SMILES string to an RDKit Mol.

    Mainly a wrapper around MolFromSmiles that raises InvalidSmiles when necessary.

    Raises:
        InvalidSmiles for empty or invalid SMILES strings.

    Args:
        smiles: SMILES string to convert.
        sanitize: whether to sanitize the molecules or not. Note: sanitization here
            corresponds to doing a sanitization with SANITIZE_ALL.
        find_radicals: usually, it will be very practical to have this set to
            True, otherwise the imported Mol instances will potentially "add"
            hydrogen atoms to radical atoms. However, in some cases it may be
            useful to de-activate it because it may cause problems on aromatic
            molecules. Irrelevant if sanitize=True.

    Returns:
        Mol instance.
    """
    mol = MolFromSmiles(smiles, sanitize=sanitize)
    if not smiles or mol is None:
        raise InvalidSmiles(smiles)

    # MolFromSmiles with sanitize=False ignores all the radicals and converts
    # them back to normal atoms. To avoid this, we need to call the sanitization
    # function, either with no sanitization, or with radical finding.
    if not sanitize:
        if find_radicals:
            sanitizations: List[SanitizeFlags | int] = [Chem.SANITIZE_FINDRADICALS]
        else:
            sanitizations = [Chem.SANITIZE_NONE]

        sanitize_mol(mol, include_sanitizations=sanitizations)

    return mol


def inchi_to_mol(inchi: str, sanitize: bool = True, removeHs: bool = True) -> Mol:
    """
    Convert an InChI string to an RDKit Mol.

    Mainly a wrapper around MolFromInchi that raises MolFromInchi when necessary.

    Raises:
        InvalidInchi for empty or invalid InChI strings.

    Args:
        inchi: InChI string to convert.
        sanitize: whether to sanitize the molecules or not. Note: sanitization here
            corresponds to doing a sanitization with SANITIZE_ALL.
        removeHs: set to True to remove Hydrogens from a molecule

    Returns:
        Mol instance.
    """
    mol: Mol = MolFromInchi(inchi, sanitize=sanitize, removeHs=removeHs)  # type: ignore[no-untyped-call]
    if not inchi or mol is None:
        raise InvalidInchi(inchi)

    return mol


def mol_to_smiles(mol: Mol, canonical: bool = True, isomericSmiles: bool = True) -> str:
    """
    Convert an RDKit Mol to a SMILES string.

    Mainly a wrapper around MolToSmiles.
    """
    return MolToSmiles(mol, canonical=canonical, isomericSmiles=isomericSmiles)


def mdl_to_mol(mdl: str, sanitize: bool = True) -> Mol:
    """
    Convert an MDL Mol string to an RDKit Mol.

    Mainly a wrapper around MolFromMolBlock that raises InvalidMdl when necessary.

    Raises:
        InvalidMdl for empty or invalid MDL mol strings.

    Args:
        mdl: MDL Mol string to convert.
        sanitize: whether to sanitize the molecules or not. Note: sanitization here
            corresponds to doing a sanitization with SANITIZE_ALL.

    Returns:
        Mol instance.
    """
    mol = MolFromMolBlock(mdl, sanitize=sanitize)
    if mol is None:
        raise InvalidMdl(mdl)

    return mol


def mol_to_mdl(mol: Mol) -> str:
    """
    Convert an RDKit Mol to an MDL Mol string.

    Mainly a wrapper around MolToMolBlock.
    """
    return MolToMolBlock(mol)


def sanitize_mol(
    mol: Mol,
    *,
    include_sanitizations: Optional[List[Union[SanitizeFlags, int]]] = None,
    exclude_sanitizations: Optional[List[Union[SanitizeFlags, int]]] = None
) -> None:
    """
    Sanitize an RDKit Mol with the specification of sanitizations to include or
    to exclude.

    Note: the RDKit sanitization function does not remove unnecessary hydrogens. See
    the function remove_hydrogens to do that.

    Raises:
        SanitizationError for unsuccessful sanitizations

    Args:
        mol: molecule to sanitize
        include_sanitizations: sanitizations to do. Exclusive with exclude_sanitizations.
        exclude_sanitizations: sanitizations to exclude, all the other ones will
            be applied. Exclusive with include_sanitizations.
    """

    # if no details about which sanitization steps to do is given, do them all
    if include_sanitizations is None and exclude_sanitizations is None:
        include_sanitizations = [Chem.SANITIZE_ALL]

    if include_sanitizations is not None and exclude_sanitizations is None:
        sanitize_ops = reduce(operator.or_, include_sanitizations, 0)
    elif include_sanitizations is None and exclude_sanitizations is not None:
        sanitize_ops = reduce(operator.xor, exclude_sanitizations, Chem.SANITIZE_ALL)
    else:
        raise ValueError(
            "Cannot specify both include_sanitizations and exclude_sanitizations."
        )

    try:
        SanitizeMol(mol, sanitizeOps=sanitize_ops)
    except Exception as e:
        raise SanitizationError(mol) from e


def mdl_to_smiles(
    mdl: str,
    sanitize: bool = True,
    canonicalize: bool = True,
    default: Optional[str] = None,
) -> str:
    """
    Convert MDL molecule description to SMILES.

    Args:
        mdl: molblock to convert
        sanitize: whether to sanitize. May be set to False to avoid RDKit valence check.
        canonicalize: whether to canonicalize for conversion to SMILES.
        default: if specified, what will be returned for conversion errors.

    Returns:
        SMILES string.
    """

    try:
        mol = mdl_to_mol(mdl, sanitize=sanitize)
    except InvalidMdl:
        if default is not None:
            return default
        else:
            raise
    return mol_to_smiles(mol, canonical=canonicalize)


def remove_hydrogens(mol: Mol) -> Mol:
    """
    Remove unnecessary hydrogens in a molecule.

    NB: The sanitization that is otherwise automatically done by RDKit is disabled.

    Returns:
        a new Mol object without unnecessary hydrogens.
    """
    return RemoveHs(mol, sanitize=False)


def canonicalize_smiles(smiles: str, check_valence: bool = True) -> str:
    """
    Canonicalize a SMILES string for a molecule.

    Args:
        smiles: SMILES string to canonicalize.
        check_valence: if False, will not do any valence check.

    Raises:
        InvalidSmiles for problems in parsing SMILES or in the sanitization.

    Returns:
        Canonicalized SMILES string.
    """
    mol = smiles_to_mol(smiles, sanitize=False, find_radicals=False)

    # NB: Removal of the unnecessary hydrogen atoms is disabled with sanitize=False above,
    # but the RDKit sanitize function does not actually do this. It is therefore
    # necessary to call this separately.
    mol = remove_hydrogens(mol)

    # Sanitization as a separate step, to enable exclusion of valence check
    try:
        excluded_sanitizations: List[SanitizeFlags | int] = []
        if not check_valence:
            excluded_sanitizations.append(Chem.SANITIZE_PROPERTIES)
        sanitize_mol(mol, exclude_sanitizations=excluded_sanitizations)
    except SanitizationError as e:
        raise InvalidSmiles(smiles) from e

    return mol_to_smiles(mol)


def maybe_canonicalize(smiles: str, check_valence: bool = True) -> str:
    """
    Canonicalize a SMILES string, but returns the original SMILES string if it fails.
    """
    try:
        return canonicalize_smiles(smiles, check_valence=check_valence)
    except InvalidSmiles:
        return smiles


def smiles_to_inchi(smiles: str, extended_tautomer_check: bool = False) -> str:
    """
    Get the InChI string for a given SMILES.

    Args:
        smiles: the SMILES to convert to InChI.
        extended_tautomer_check: include the options for additional tautomer standardization.

    Raises:
        InvalidSmiles for conversion errors or invalid SMILES.

    Returns:
        InChI string.
    """

    mol = smiles_to_mol(smiles, sanitize=False)

    # Due to a bug (?) in RDKit, it is necessary to reassign the stereochemistry
    # before conversion to InChi: https://github.com/rdkit/rdkit/issues/2361.
    AssignStereochemistry(mol, cleanIt=True, force=True)

    try:
        return mol_to_inchi(mol, extended_tautomer_check=extended_tautomer_check)
    except Exception:
        raise InvalidSmiles(smiles)


def mol_to_inchi(mol: Mol, extended_tautomer_check: bool = False) -> str:
    """
    Convert an RDKit Mol to an InChI.

    Args:
        mol: the RDKit Mol to convert to InChI.
        extended_tautomer_check: include the options for additional tautomer
            standardization.

    Returns:
        InChI string.
    """

    options = ""
    if extended_tautomer_check:
        # https://pubs.acs.org/doi/10.1021/acs.jcim.9b01080
        options = "-KET -15T"

    return Chem.MolToInchi(mol, options=options)  # type: ignore


def cleanup_smiles(smiles: str) -> str:
    """
    Cleanup a SMILES string, doing the bare minimum.

    This means that no canonicalization, no valence check, no kekulization, etc,
    will be done.
    See the unit tests for examples.

    A minimal sanitization (SANITIZE_FINDRADICALS) is necessary, otherwise
    "[C]" is converted to "C".

    Args:
        smiles: SMILES to clean up.

    Returns:
        A cleaned-up SMILES string.
    """
    mol = smiles_to_mol(smiles, sanitize=False)
    sanitize_mol(mol, include_sanitizations=[Chem.SANITIZE_FINDRADICALS])
    return mol_to_smiles(mol, canonical=False)


def split_smiles_and_fragment_info(reaction_smiles: str) -> Tuple[str, str]:
    """
    The reaction SMILES from Pistachio sometimes contain fraction information
    at the end of the given string. This function splits both parts of the
    reaction SMILES.

    Args:
        reaction_smiles: (potentially extended) reaction SMILES.

    Returns:
        Tuple: ('pure' reaction SMILES, fragment information).
    """

    m = re.search(r"^(\S+) ?(.*)$", reaction_smiles)
    assert m is not None
    return m.group(1), m.group(2)



import re
from typing import Any, Dict, Generator, List, Tuple

from rxn.utilities.regex import capturing

_LIST_REGEX_STRING = capturing(".+") + r"\(" + capturing(r"\d+") + r"\)"
_LIST_REGEX = re.compile(_LIST_REGEX_STRING)


class PropertyParser:
    """
    To parse properties given in a RDF / MDL into a nested dictionary.
    """

    def __init__(self) -> None:
        self.result: Dict[str, Any] = {}

    def parse_dict(self, property_dict: Dict[str, str]) -> None:
        for key, value in property_dict.items():
            self.parse_property(key, value)

    def parse_property(self, key: str, value: str) -> None:
        self._parse_property(self.result, key, value)

    def _is_list_property(self, subkey: str) -> bool:
        return _LIST_REGEX.match(subkey) is not None

    def _parse_property(self, container: Dict[str, Any], key: str, value: str) -> None:
        if not key:
            raise ValueError("A key must be provided.")

        splits = key.split(":")
        if len(splits) == 1:
            container[key] = value
            return

        if self._is_list_property(splits[0]):
            self._parse_list_property(container, splits, value)
        else:
            self._parse_dict_property(container, splits, value)

    def _parse_list_property(
        self, container: Dict[str, Any], key_splits: List[str], value: str
    ) -> None:
        list_match = _LIST_REGEX.match(key_splits[0])
        if list_match is None:
            raise RuntimeError(
                "Not a list property - by construction, this should not happen."
            )

        subkey = list_match.group(1)
        list_index = int(list_match.group(2))

        # Initialize list if needed
        if subkey not in container:
            container[subkey] = []

        subkey_list = container[subkey]

        # Add new dicts if necessary
        for _ in range(len(subkey_list), list_index):
            subkey_list.append({})

        self._parse_property(
            subkey_list[list_index - 1], ":".join(key_splits[1:]), value
        )

    def _parse_dict_property(
        self, container: Dict[str, Any], key_splits: List[str], value: str
    ) -> None:
        subkey = key_splits[0]
        if subkey not in container:
            container[subkey] = {}
        self._parse_property(container[subkey], ":".join(key_splits[1:]), value)


class PropertySerializer:
    """Do the reverse operation compared to PropertyParser."""

    def convert_dict(self, container: Dict[str, Any]) -> Dict[str, str]:
        return {
            key: value
            for key, value in self._convert_dict(prefix="", container=container)
        }

    def _convert_dict(
        self, prefix: str, container: Dict[str, Any]
    ) -> Generator[Tuple[str, str], None, None]:
        for key, value in container.items():
            yield from self._convert(prefix=prefix, key=key, current=value)

    def _convert(
        self, prefix: str, key: str, current: Any
    ) -> Generator[Tuple[str, str], None, None]:
        if isinstance(current, str):
            yield f"{prefix}{key}", current
        elif isinstance(current, list):
            for index, v in enumerate(current, 1):
                yield from self._convert_dict(
                    prefix=f"{prefix}{key}({index}):", container=v
                )
        elif isinstance(current, dict):
            yield from self._convert_dict(prefix=f"{prefix}{key}:", container=current)
        else:
            raise RuntimeError(f"Not supported for property serialization: {current}")


def parse_properties(properties: Dict[str, str]) -> Dict[str, Any]:
    """Parse the properties given in the RDF into a nested dictionary."""
    pp = PropertyParser()
    pp.parse_dict(properties)
    return pp.result


def serialize_properties(properties: Dict[str, Any]) -> Dict[str, str]:
    """Do the reverse operation compared to parse_properties."""
    return PropertySerializer().convert_dict(properties)



import logging
from typing import Optional

from .conversion import inchi_to_mol, mol_to_inchi, mol_to_smiles, smiles_to_mol
from .exceptions import InvalidInchi, InvalidSmiles

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

RXN_SMILES_SEPARATOR = ">>"


def standardize_smiles(
    smiles: str,
    canonicalize: bool = True,
    sanitize: bool = True,
    find_radicals: bool = True,
    inchify: bool = False,
) -> str:
    """Ensure that a SMILES follows a desired standard.

    It allows canonicalization, sanitization and inchification keeping stereochemistry with isomericSmile=True.
    It can process multiple molecules separated by ".".
    Note that inchify set to True will also canonicalize the molecule.

    Args:
        smiles: SMILES representation of a molecule.
        canonicalize: canonicalize SMILES. Defaults to True.
        sanitize: sanitize SMILES. Defaults to True.
        inchify: inchify the SMILES. Defaults to False.

    Returns:
        a SMILES following the desired standard.
    """
    try:
        molecule = smiles_to_mol(smiles, sanitize=sanitize, find_radicals=find_radicals)
    except InvalidSmiles:
        logger.error(f"SMILES parsing failure: {smiles}.")
        raise

    if inchify:
        try:
            inchi_string = mol_to_inchi(molecule)
        except InvalidInchi:
            logger.error(
                f"Inchification failure for SMILES: {smiles}. Returning its canonical version."
            )
            return mol_to_smiles(molecule, isomericSmiles=True)
        else:
            # canonical set to True because we can't guarantee no canonicalization
            try:
                molecule_from_inchi = inchi_to_mol(inchi_string)
            except InvalidInchi:
                logger.error(
                    f"De-inchification failure for InChi: {inchi_string}. Returning its canonical version."
                )
                return mol_to_smiles(molecule, isomericSmiles=True)
            return mol_to_smiles(molecule_from_inchi, canonical=True)
    if canonicalize:
        return mol_to_smiles(molecule, isomericSmiles=True)
    else:
        return smiles


def standardize_molecules(
    molecules: str,
    canonicalize: bool = True,
    sanitize: bool = True,
    inchify: bool = False,
    fragment_bond: str = "~",
    ordered_precursors: bool = True,
    molecule_token_delimiter: Optional[str] = None,
    is_enzymatic: bool = False,
    enzyme_separator: str = "|",
) -> str:
    """Ensure that a set of molecules represented by a string follows a desired standard.

    Args:
        molecules: molecules SMILES. Molecules can be separated via a ".".
            Fragments are supported with a custom `fragment_bond`.
        canonicalize: canonicalize SMILES. Defaults to True.
        sanitize: sanitize SMILES. Defaults to True.
        inchify: inchify the SMILES. Defaults to False.
        fragment_bond: fragment bond. Defaults to '~'.
        ordered_precursors: order precursors. Defaults to True.
        molecule_token_delimiter: delimiter for big molecule tokens. Defaults to None
        is_enzymatic: the molecules are representing an enzymatic reaction. Defaults to False.
        enzyme_separator: separator for molecules and the enzyme. Defaults to '|'.

    Returns:
        standardized molecules.

    Examples:
        Standardize multiple molecules:
        >>> standardize_molecules('CCO.CC')
        'CC.CCO'
        Standardize multiple molecules including fragment information:
        >>> standardize_molecules('CCO.CC~C')
        'CCO.C~CC'
    """
    enzyme = ""
    if is_enzymatic:
        splitted_molecules = molecules.split(enzyme_separator)
        molecules = splitted_molecules[0]
        if len(splitted_molecules) > 1:
            enzyme = splitted_molecules[1]
            enzyme = "{}{}".format(enzyme_separator, enzyme)
    if molecule_token_delimiter is not None:
        molecules = molecules.replace(molecule_token_delimiter, "")
    if fragment_bond in molecules:
        standardized_molecules_list = [
            # make sure we remove the fragment to have valid SMILES
            standardize_smiles(
                molecule.replace(fragment_bond, "."),
                canonicalize=canonicalize,
                sanitize=sanitize,
                inchify=inchify,
            ).replace(".", fragment_bond)
            for molecule in molecules.split(".")
        ]
        if ordered_precursors:
            standardized_molecules_list = sorted(standardized_molecules_list)
        standardized_molecules = ".".join(standardized_molecules_list)
    else:
        if ordered_precursors:
            # RDKit guarantees ordered precursors
            standardized_molecules = standardize_smiles(
                molecules,
                canonicalize=canonicalize,
                sanitize=sanitize,
                inchify=inchify,
            )
        else:
            standardized_molecules_list = [
                standardize_smiles(
                    molecule,
                    canonicalize=canonicalize,
                    sanitize=sanitize,
                    inchify=inchify,
                )
                for molecule in molecules.split(".")
            ]
            standardized_molecules = ".".join(standardized_molecules_list)
    # add optional enzyme information
    standardized_molecules = "{}{}".format(standardized_molecules, enzyme)
    return standardized_molecules



from rxn.chemutils.utils import remove_atom_mapping


def test_remove_atom_mapping() -> None:
    # Reaction with mapping
    with_mapping = (
        "[*:1]CCO.[CH3:1][C:2](=[O:3])[OH:4]>[H+]>CC[O:4][C:2](=[O:3])[CH3:1].O"
    )
    without_mapping = "[*:1]CCO.[CH3][C](=[O])[OH]>[H+]>CC[O][C](=[O])[CH3].O"
    assert remove_atom_mapping(with_mapping) == without_mapping

    # Reaction with mapping -> do nothing
    rxn_without_mapping = "[*:1]CCO.CC(=O)O>[H+]>CCOC(C)=O.O"
    assert remove_atom_mapping(rxn_without_mapping) == rxn_without_mapping

    # Reaction with fragment info -> fragment info is not removed
    fragment_with_mapping = (
        "[*:1][Cl-].[Cl-].[Cl-].C(Cl)Cl>[CH3:9][CH:8]([CH3:10])[c:7]1[cH:11][cH:12][c:13]"
        "([cH:14][cH:15]1)[C:5](=[O:6])[CH2:4][CH2:3][CH2:2][Cl:1] |f:2.3.4.5|"
    )
    fragment_without_mapping = (
        "[*:1][Cl-].[Cl-].[Cl-].C(Cl)Cl>[CH3][CH]([CH3])[c]1[cH][cH][c]"
        "([cH][cH]1)[C](=[O])[CH2][CH2][CH2][Cl] |f:2.3.4.5|"
    )
    assert remove_atom_mapping(fragment_with_mapping) == fragment_without_mapping



import pytest

from rxn.chemutils.exceptions import InvalidSmiles
from rxn.chemutils.multicomponent_smiles import (
    apply_to_multicomponent_smiles,
    canonicalize_multicomponent_smiles,
    list_to_multicomponent_smiles,
    multicomponent_smiles_to_list,
    remove_duplicates_in_multicomponent_smiles,
    sort_multicomponent_smiles,
)


def test_multicomponent_smiles_to_list() -> None:
    assert multicomponent_smiles_to_list("") == []
    assert multicomponent_smiles_to_list("A.B.C") == ["A", "B", "C"]
    assert multicomponent_smiles_to_list("..A.B.C.") == ["A", "B", "C"]
    assert multicomponent_smiles_to_list("A~B.C") == ["A~B", "C"]
    assert multicomponent_smiles_to_list("A~B.C", fragment_bond="~") == ["A.B", "C"]
    assert multicomponent_smiles_to_list("A#B.C", fragment_bond="#") == ["A.B", "C"]


def test_list_to_multicomponent_smiles() -> None:
    assert list_to_multicomponent_smiles([]) == ""
    assert list_to_multicomponent_smiles(["A", "B", "C"]) == "A.B.C"
    assert list_to_multicomponent_smiles(["A.B", "C"]) == "A.B.C"
    assert list_to_multicomponent_smiles(["A.B", "C"], fragment_bond="~") == "A~B.C"
    assert list_to_multicomponent_smiles(["A.B", "C"], fragment_bond="#") == "A#B.C"


def test_apply_to_multicomponent_smiles() -> None:
    def dummy(smiles: str) -> str:
        """Dummy function that appends the number of fragments in a SMILES string to it."""
        return smiles + str(smiles.count(".") + 1)

    assert apply_to_multicomponent_smiles("A.B.CDE", dummy) == "A1.B1.CDE1"
    assert apply_to_multicomponent_smiles("A.B~C", dummy) == "A1.B~C1"
    assert (
        apply_to_multicomponent_smiles("A.B~C", dummy, fragment_bond="~") == "A1.B~C2"
    )
    assert (
        apply_to_multicomponent_smiles("ABCD.B~C~D", dummy, fragment_bond="~")
        == "ABCD1.B~C~D3"
    )


def test_canonicalize_multicomponent_smiles() -> None:
    # A few basic examples
    assert canonicalize_multicomponent_smiles("C(C)O") == "CCO"
    assert canonicalize_multicomponent_smiles("C.OCC.C(C)O") == "C.CCO.CCO"

    # Canonicalizing an empty string is valid - since this is a valid multi-component SMILES
    assert canonicalize_multicomponent_smiles("") == ""

    # with fragment bond; may fail if fragment bond not specified
    assert (
        canonicalize_multicomponent_smiles("C.[H-]~[F-]~[Mg+2].O", fragment_bond="~")
        == "C.[F-]~[H-]~[Mg+2].O"
    )
    with pytest.raises(InvalidSmiles):
        _ = canonicalize_multicomponent_smiles("C.[H-]~[F-]~[Mg+2].O")

    # possibility to check valence or not
    assert (
        canonicalize_multicomponent_smiles("C.CF(C).O", check_valence=False)
        == "C.CFC.O"
    )
    with pytest.raises(InvalidSmiles):
        _ = canonicalize_multicomponent_smiles("C.CF(C).O")


def test_sort_multicomponent_smiles() -> None:
    assert sort_multicomponent_smiles("") == ""
    assert sort_multicomponent_smiles("B.D.C.A") == "A.B.C.D"
    assert sort_multicomponent_smiles("B.D.C~A") == "B.C~A.D"
    assert sort_multicomponent_smiles("NCC(N).CCO.COCO") == "CCO.COCO.NCC(N)"


def test_remove_duplicates_in_multicomponent_smiles() -> None:
    assert remove_duplicates_in_multicomponent_smiles("B.D.A.E") == "B.D.A.E"
    assert remove_duplicates_in_multicomponent_smiles("B.D.A.A.D.E") == "B.D.A.E"
    assert remove_duplicates_in_multicomponent_smiles("CC.C.O.N.O") == "CC.C.O.N"

    # Note: does not remove compounds that are the same when canonicalized!
    assert (
        remove_duplicates_in_multicomponent_smiles("CC.C(C).(C)(C)") == "CC.C(C).(C)(C)"
    )



import pytest
from rxn.utilities.files import (
    dump_list_to_file,
    load_list_from_file,
    named_temporary_path,
)

from rxn.chemutils.exceptions import UnclearWhetherTokenized
from rxn.chemutils.tokenization import (
    TokenizationError,
    detokenize_file,
    detokenize_smiles,
    ensure_tokenized_file,
    file_is_tokenized,
    string_is_tokenized,
    to_tokens,
    tokenize_file,
    tokenize_smiles,
)


def test_tokenization_error() -> None:
    error = TokenizationError("ATokenizationError", "my message")
    assert error.type == "TokenizationError"
    assert error.title == "ATokenizationError"
    assert error.detail == "my message"
    with pytest.raises(ValueError):
        raise error


def test_to_tokens() -> None:
    smiles = r"C\C=C\C"
    groundtruth = ["C", "\\", "C", "=", "C", "\\", "C"]
    tokens = to_tokens(smiles)
    assert tokens == groundtruth
    smiles = "CCZCC"
    with pytest.raises(TokenizationError):
        try:
            to_tokens(smiles)
        except TokenizationError as exception:
            assert exception.title == "SmilesJoinedTokensMismatch"
            assert exception.detail == 'SMILES="CCZCC" != joined_tokens="CCCC"'
            raise


def test_tokenize_simple_smiles() -> None:
    # For this simple SMILES, 1 letter per token
    smiles = "CN1C=NC2=C1C(=O)N(C(=O)N2C)C"
    expected = "C N 1 C = N C 2 = C 1 C ( = O ) N ( C ( = O ) N 2 C ) C"
    assert tokenize_smiles(smiles) == expected


def test_tokenize_complex_smiles() -> None:
    # Some tokens have more than one letter
    smiles = "C([N+](=O)[O-])1=CC(Br)=CC=C1"
    expected = "C ( [N+] ( = O ) [O-] ) 1 = C C ( Br ) = C C = C 1"
    assert tokenize_smiles(smiles) == expected


def test_tokenize_two_digit_cycle() -> None:
    smiles = "C%42CCCCC%42"
    expected = "C %42 C C C C C %42"
    assert tokenize_smiles(smiles) == expected


def test_tokenize_three_digit_cycle() -> None:
    smiles = "C%(432)CCCCC%(432)"
    expected = "C %(432) C C C C C %(432)"
    assert tokenize_smiles(smiles) == expected


def test_tokenize_reaction_smiles() -> None:
    smiles = "COCO.OCC>>NCOC"
    expected = "C O C O . O C C >> N C O C"
    assert tokenize_smiles(smiles) == expected


def test_tokenize_reaction_smiles_with_fragment_bond() -> None:
    smiles = "COCO.[Na+]~[OH-].OCC>O.C>NCOC"
    expected = "C O C O . [Na+] ~ [OH-] . O C C > O . C > N C O C"
    assert tokenize_smiles(smiles) == expected


def test_tokenize_with_fallback_value() -> None:
    assert tokenize_smiles("CCOC", "fallback") == "C C O C"
    assert tokenize_smiles("C C O C", "fallback") == "fallback"
    assert tokenize_smiles("invalid", "fallback") == "fallback"


def test_detokenize() -> None:
    # Basically, detokenizing simply removes the spaces
    tokenized = "some random string also not S M I L E S ."
    expected = "somerandomstringalsonotSMILES."
    assert detokenize_smiles(tokenized) == expected


def test_tokenization_roundtrip() -> None:
    smiles_list = [
        "CN1C=NC2=C1C(=O)N(C(=O)N2C)C",  # simple SMILES
        "C([N+](=O)[O-])1=CC(Br)=CC=C1",  # more complex SMILES
        "C%42CCCCC%42",  # SMILES with two-digit cycle
        "C%(432)CCCCC%(432)",  # SMILES with three-digit cycle
        "COCO.OCC>>NCOC",  # reaction SMILES
        "COCO.[Na+]~[OH-].OCC>O.C>NCOC",  # reaction SMILES with fragment bonds
    ]

    for smiles in smiles_list:
        assert detokenize_smiles(tokenize_smiles(smiles)) == smiles


def test_string_is_tokenized() -> None:
    assert string_is_tokenized("C C O . [Na+]")
    assert string_is_tokenized("C C O . [Na+] >> C ( O ) Cl")
    assert not string_is_tokenized("C C O . [Na+] >> C (O) Cl")
    assert not string_is_tokenized("CCO")

    # Special cases - empty strings and strings with single tokens
    for string in ["", "C", ">>", "[Na]"]:
        with pytest.raises(UnclearWhetherTokenized):
            _ = string_is_tokenized(string)

    # Tokenization errors are being propagated
    with pytest.raises(TokenizationError):
        _ = string_is_tokenized("I N V A L I D")


def test_file_is_tokenized() -> None:
    # Basic tokenized example
    with named_temporary_path() as path:
        dump_list_to_file(["C C O >> C C O", "C C . C"], path)
        assert file_is_tokenized(path)

    # Basic non-tokenized example
    with named_temporary_path() as path:
        dump_list_to_file(["CCO>>CCO", "CC.C"], path)
        assert not file_is_tokenized(path)

    # Only checks the first line - returns True even if the second line is not tokenized
    with named_temporary_path() as path:
        dump_list_to_file(["C C O >> C C O", "CC.C"], path)
        assert file_is_tokenized(path)

    # empty file
    with named_temporary_path() as path:
        dump_list_to_file([], path)
        with pytest.raises(RuntimeError):
            _ = file_is_tokenized(path)

    # Invalid SMILES
    with named_temporary_path() as path:
        dump_list_to_file(["I N V A L I D", "CC.C"], path)
        with pytest.raises(TokenizationError):
            _ = file_is_tokenized(path)

    # Empty first line - needs to check the second line!
    with named_temporary_path() as path:
        dump_list_to_file(["", "C C O >> C C O"], path)
        assert file_is_tokenized(path)
    with named_temporary_path() as path:
        dump_list_to_file(["", "CCO>>CCO"], path)
        assert not file_is_tokenized(path)

    # First line has one single token - needs to check the second line!
    with named_temporary_path() as path:
        dump_list_to_file([">>", "C C O >> C C O"], path)
        assert file_is_tokenized(path)
    with named_temporary_path() as path:
        dump_list_to_file([">>", "CCO>>CCO"], path)
        assert not file_is_tokenized(path)


def test_tokenize_file() -> None:
    with named_temporary_path() as f_in, named_temporary_path() as f_out:
        # Original content
        original = ["CCO>>CCO", "CC.C", "INVALID", "C(NCC)[S]OC"]
        dump_list_to_file(original, f_in)

        # Expected (tokenized) content
        fallback = "ERROR"
        tokenized = ["C C O >> C C O", "C C . C", fallback, "C ( N C C ) [S] O C"]

        tokenize_file(f_in, f_out, fallback_value=fallback)

        assert load_list_from_file(f_out) == tokenized


def test_detokenize_file() -> None:
    with named_temporary_path() as f_in, named_temporary_path() as f_out:
        # Original (tokenized) content
        original = ["C C O >> C C O", "C C . C", "C ( N C C ) [S] O C"]
        dump_list_to_file(original, f_in)

        # Expected (detokenized) content
        detokenized = ["CCO>>CCO", "CC.C", "C(NCC)[S]OC"]

        detokenize_file(f_in, f_out)

        assert load_list_from_file(f_out) == detokenized


def test_ensure_tokenized_file() -> None:
    with named_temporary_path() as temporary_path:
        temporary_path.mkdir()

        # prepare filenames
        postfix = ".tknz"
        already_tokenized_file = str(temporary_path / "a.txt")
        not_tokenized_file = str(temporary_path / "b.txt")
        updated_tokenized_file = str(temporary_path / "b.txt") + postfix

        # contents (original and expected)
        fallback = "error"
        tokenized = ["C C O >> C C O", "C C . C", "C ( N C C ) [S] O"]
        not_tokenized = ["CCO>>CCO", "CC.C", "INVALID", "C(N)[S]O"]
        after_tokenization = ["C C O >> C C O", "C C . C", fallback, "C ( N ) [S] O"]

        # Put into files
        dump_list_to_file(tokenized, already_tokenized_file)
        dump_list_to_file(not_tokenized, not_tokenized_file)

        # ensure for already tokenized - the original unchanged file can be used
        result = ensure_tokenized_file(
            already_tokenized_file, postfix=postfix, fallback_value=fallback
        )
        assert result == already_tokenized_file
        assert load_list_from_file(result) == tokenized

        # ensure for non-tokenized - a new file was created with tokenized strings
        result = ensure_tokenized_file(
            not_tokenized_file, postfix=postfix, fallback_value=fallback
        )
        assert result == updated_tokenized_file
        assert load_list_from_file(updated_tokenized_file) == after_tokenization



import re
from typing import List, Tuple

from .conversion import split_smiles_and_fragment_info
from .reaction_equation import ReactionEquation, cleanup_compounds
from .utils import remove_atom_mapping

# Regex pattern to extract the fragment info from the extended info of reaction SMILES
EXTENDED_FRAGMENT_REGEX = re.compile(r"f:[\d\.,]+")

# Regex pattern to extract the fragment groups from the fragment info
FRAGMENT_GROUP_REGEX = re.compile(r"(\d+(?:\.\d+)*)")


class UnsupportedExtendedReactionSmiles(ValueError):
    def __init__(self, reaction_smiles: str):
        super().__init__(
            f'The syntax of "{reaction_smiles}" is not supported by RDKit.'
        )


def parse_extended_reaction_smiles(
    extended_reaction_smiles: str, remove_atom_maps: bool = True
) -> ReactionEquation:
    """
    Convert an extended reaction SMILES (with potential fragment information)
    to a ReactionEquation instance.

    Args:
        extended_reaction_smiles: extended reaction SMILES
        remove_atom_maps: whether to remove the atom mapping information.

    Returns:
        ReactionEquation instance
    """
    return _Importer.convert(
        extended_reaction_smiles, remove_atom_maps=remove_atom_maps
    )


def to_extended_reaction_smiles(reaction: ReactionEquation) -> str:
    """
    Convert a ReactionEquation instance to an extended reaction SMILES (with
    potential fragment information).

    Args:
        reaction: reaction equation to convert

    Returns:
        The extended reaction SMILES string.
    """
    return _Exporter.convert(reaction)


class _Importer:
    """
    Convert extended reaction SMILES to ReactionEquation instances.
    """

    @staticmethod
    def convert(
        extended_reaction_smiles: str, remove_atom_maps: bool
    ) -> ReactionEquation:
        """
        Convert an extended SMILES to a reaction equation.

        Used to rely more on RDKit; now does as few RDKit operations as necessary.

        Args:
            extended_reaction_smiles: extended reaction SMILES
            remove_atom_maps: whether to remove the atom mapping information.
        """

        pure_smiles, fragment_info = split_smiles_and_fragment_info(
            extended_reaction_smiles
        )

        if remove_atom_maps:
            pure_smiles = remove_atom_mapping(pure_smiles)

        # We split at the ">" characters, only if they are not preceded by a "-",
        # which would indicate a dative bond.
        reactant_groups = re.split(r"(?<!-)>", pure_smiles)
        mols_groups = [group.split(".") for group in reactant_groups]
        mols_groups = [[mol for mol in group if mol] for group in mols_groups]

        fragment_groups = determine_fragment_groups(fragment_info)
        groups = _Importer.group_fragments(mols_groups, fragment_groups)

        reaction_equation = ReactionEquation(*groups)

        # If the atom maps were removed, cleaning up the compounds will
        # return a much nicer string.
        if remove_atom_maps:
            reaction_equation = cleanup_compounds(reaction_equation)

        return reaction_equation

    @staticmethod
    def group_fragments(
        raw_smiles_groups: List[List[str]], fragment_groups: List[List[int]]
    ) -> List[List[str]]:
        """
        Merge the reaction fragments belonging together.
        """
        merged_groups = []

        offset = 0
        for raw_smiles_group in raw_smiles_groups:
            merged_smiles_group = merge_molecules_from_fragment_groups(
                raw_smiles_group, fragment_groups, offset
            )
            merged_groups.append(merged_smiles_group)
            offset += len(raw_smiles_group)

        return merged_groups


class _Exporter:
    """
    Convert ReactionEquation to extended reaction SMILES with fragment information.
    """

    @staticmethod
    def convert(reaction: ReactionEquation) -> str:
        offset = 0
        reactants, reactant_groups = _Exporter.fragment_group(
            reaction.reactants, offset
        )
        offset += len(reactants)
        agents, agent_groups = _Exporter.fragment_group(reaction.agents, offset)
        offset += len(agents)
        products, product_groups = _Exporter.fragment_group(reaction.products, offset)

        groups = reactant_groups + agent_groups + product_groups

        smiles_groups = (
            ".".join(smiles for smiles in group)
            for group in (reactants, agents, products)
        )
        smiles_without_fragment_info = ">".join(smiles_groups)

        if not groups:
            return smiles_without_fragment_info

        fragment_info = _Exporter.generate_fragment_info(groups)
        return f"{smiles_without_fragment_info} {fragment_info}"

    @staticmethod
    def fragment_group(
        compounds: List[str], offset: int
    ) -> Tuple[List[str], List[List[int]]]:
        """
        Converts a group of molecules, some of which possibly are composed of several fragments,
        to the list of SMILES to put in the final reaction SMILES, and a list of groups of molecules
        belonging together.

        Example:
            ['O', '[Na+].[OH-]'] -> (['O', '[Na+]', '[OH-]'], [[1, 2]])

        Args:
            compounds: SMILES string for molecules of the same group (f.i. reactants)
            offset: index, in the final reaction SMILES, of the first molecule of that group
        """

        smiles_list: List[str] = []
        groups: List[List[int]] = []

        current_index = offset
        for c in compounds:
            molecules = c.split(".")
            smiles_list.extend(molecules)

            number_fragments = len(molecules)
            if number_fragments > 1:
                groups.append(
                    list(range(current_index, current_index + number_fragments))
                )

            current_index += number_fragments

        return smiles_list, groups

    @staticmethod
    def generate_fragment_info(groups: List[List[int]]) -> str:
        if not groups:
            return ""

        group_strings = [".".join(str(number) for number in g) for g in groups]
        all_groups = ",".join(group_strings)
        return f"|f:{all_groups}|"


def determine_fragment_groups(extended_reaction_info: str) -> List[List[int]]:
    """
    From the fragment info string (such as ``|f:0.2,5.6|``), determine the groups
    of indices that must be grouped.

    Args:
        extended_reaction_info: Extended reaction info, potentially containing
            information about fragments, stereochemistry, etc. See documentation on
            https://docs.chemaxon.com/display/docs/chemaxon-extended-smiles-and-smarts-cxsmiles-and-cxsmarts.md

    Returns:
        List of groups (f.i. [[0,2], [5,6]])
    """

    # Extract the part related to fragments, starting with "f:"
    fragment_subpart_match = EXTENDED_FRAGMENT_REGEX.search(extended_reaction_info)
    if fragment_subpart_match is None:
        return []
    fragment_info = fragment_subpart_match.group(0)

    # Extract the groups of associated numbers as strings
    fragment_group_matches = FRAGMENT_GROUP_REGEX.findall(fragment_info)

    # Convert to lists of integers
    groups = []
    for match in fragment_group_matches:
        indices = match.split(".")
        groups.append([int(i) for i in indices])
    return groups


def merge_molecules_from_fragment_groups(
    smiles_list: List[str], fragment_groups: List[List[int]], offset: int
) -> List[str]:
    """
    Combine molecules according to the fragment definition.

    Args:
        smiles_list: SMILES strings to potentially merge
        fragment_groups: indices for groups of molecules belonging together; some may be out of range
        offset: what index the first molecule in smiles_list corresponds to

    Returns:
        List of molecules, some of which were potentially merged. Non-merged molecules come first.
    """

    allowed_indices = set(range(len(smiles_list)))
    merged_indices = set()

    merged_molecules = []

    for group in fragment_groups:
        relative_indices = [i - offset for i in group]
        in_range = [e in allowed_indices for e in relative_indices]

        # either all indices are in the range, or none is
        all_in_range = all(in_range)
        none_in_range = not any(in_range)
        if not (all_in_range or none_in_range):
            raise ValueError()

        if all_in_range:
            merged_molecule = ".".join(smiles_list[i] for i in relative_indices)
            merged_molecules.append(merged_molecule)
            merged_indices.update(relative_indices)

    remaining_molecule_indices = sorted(list(allowed_indices - merged_indices))
    unmerged_molecules = [smiles_list[i] for i in remaining_molecule_indices]

    return unmerged_molecules + merged_molecules



import click
from rxn.utilities.files import load_list_from_file

from rxn.chemutils.reaction_combiner import ReactionCombiner
from rxn.chemutils.reaction_smiles import ReactionFormat


@click.command()
@click.argument("fragments_1_file", type=click.Path(exists=True, dir_okay=False))
@click.argument("fragments_2_file", type=click.Path(exists=True, dir_okay=False))
@click.option("--standardize/--no-standardize", default=False)
@click.option(
    "--reaction_format",
    type=click.Choice(
        ["standard", "standard_with_tilde", "extended"], case_sensitive=False
    ),
    default="extended",
)
def main(
    fragments_1_file: str,
    fragments_2_file: str,
    standardize: bool,
    reaction_format: str,
) -> None:
    """Combine precursors and products (or two sets of partial reactions, and
    write the reactions into std output.

    If one of both file sizes is a multiple of the other, we consider it to
    have been generated via a "top-N" prediction.
    """
    fragments_1 = load_list_from_file(fragments_1_file)
    fragments_2 = load_list_from_file(fragments_2_file)

    combiner = ReactionCombiner(
        standardize=standardize,
        reaction_format=ReactionFormat.from_string(reaction_format),
    )

    for reaction_smiles in combiner.combine(fragments_1, fragments_2):
        print(reaction_smiles)


if __name__ == "__main__":
    main()



import re
from pathlib import Path
from typing import Callable, Iterator, List, Optional, Union

from rxn.utilities.regex import capturing

from .rdf_reaction import RdfReaction


class RdfParsingError(RuntimeError):
    """Exception for RDF parsing errors."""


class InvalidBlock(RdfParsingError):
    """Exception raised when a block of an RDF file cannot be processed."""

    def __init__(self, lines: List[str]):
        self.lines = lines
        super().__init__("Invalid block:\n" + "\n".join(lines))


class IncompleteReaction(RdfParsingError):
    """Exception for incomplete reaction in RDF file."""


BLOCK_TYPE_REGEX = re.compile(r"\$" + capturing(r"\w+"))
RIREG_REGEX = re.compile(r"\$RFMT \$[RM]IREG " + capturing(r"\d+"))
DTYPE_REGEX = re.compile(r"\$DTYPE " + capturing(".*"))
DATUM_REGEX = re.compile(r"\$DATUM " + capturing(".*"))


class ParsedReaction:
    """
    Reaction under construction during parsing of an RDF file.
    """

    def __init__(self) -> None:
        self.rireg: Optional[int] = None
        self.n_precursors: Optional[int] = None
        self.n_products: Optional[int] = None
        self.mols: List[str] = []
        self.dtypes: List[str] = []
        self.datums: List[str] = []

    def handle_line_block(self, lines: List[str]) -> None:
        block_type = self.block_type(lines)
        if block_type == "RFMT":
            self.handle_rfmt(lines)
        elif block_type == "RXN":
            self.handle_rxn(lines)
        elif block_type == "MOL":
            self.handle_mol(lines)
        elif block_type == "DTYPE":
            self.handle_dtype(lines)
        elif block_type == "DATUM":
            self.handle_datum(lines)
        else:
            raise ValueError(f"Invalid block type: {block_type}")

    def handle_rfmt(self, lines: List[str]) -> None:
        if len(lines) != 1:
            raise InvalidBlock(lines)
        match = RIREG_REGEX.match(lines[0])
        if match is None:
            raise InvalidBlock(lines)
        self.rireg = int(match.group(1))

    def handle_rxn(self, lines: List[str]) -> None:
        if len(lines) != 5:
            raise InvalidBlock(lines)

        self.n_precursors, self.n_products = [int(x) for x in lines[4].split()]

    def handle_mol(self, lines: List[str]) -> None:
        self.mols.append("\n".join(lines[1:]))

    def handle_dtype(self, lines: List[str]) -> None:
        if len(lines) != 1:
            raise InvalidBlock(lines)

        match = DTYPE_REGEX.match(lines[0])
        if match is None:
            raise InvalidBlock(lines)
        self.dtypes.append(match.group(1))

    def handle_datum(self, lines: List[str]) -> None:
        match = DATUM_REGEX.match(lines[0])
        if match is None:
            raise InvalidBlock(lines)
        lines[0] = match.group(1)

        # special case: leave out $MFMT, otherwise the property will not be
        # valid MolBlocks for parsing with RDKit.
        if lines[0] == "$MFMT":
            lines = lines[1:]

        self.datums.append("\n".join(lines))

    def block_type(self, lines: List[str]) -> str:
        """Get the type of a block: RXN, DTYPE, DATUM, etc."""
        match = BLOCK_TYPE_REGEX.match(lines[0])
        if match is None:
            raise InvalidBlock(lines)
        return match.group(1)

    def to_reaction(self) -> RdfReaction:
        if self.rireg is None or self.n_precursors is None or self.n_products is None:
            raise IncompleteReaction()

        if len(self.mols) != self.n_precursors + self.n_products:
            raise RdfParsingError()

        precursors = self.mols[: self.n_precursors]
        products = self.mols[-self.n_products :]

        if len(self.dtypes) != len(self.datums):
            raise RdfParsingError()

        meta = {key: value for key, value in zip(self.dtypes, self.datums)}

        return RdfReaction(
            reactants=precursors,
            reagents=[],
            products=products,
            meta=meta,
            reaction_index=self.rireg,
        )


class RdfParser:
    """
    Custom parser for RDF files.
    """

    def __init__(self, filename: Union[Path, str], encoding: str = "latin-1"):
        """
        Args:
            filename: path to the RDF file to read.
            encoding: file encoding. Defaults to latin-1 because Thieme has such
                an encoding for several files.
        """
        self.filename = filename
        self.encoding = encoding

    def __iter__(self) -> Iterator[RdfReaction]:
        yield from self.iter_reactions()

    def iter_reactions(self) -> Iterator[RdfReaction]:
        block_iterator = self.iter_blocks()
        # Consume line with RDFILE
        _ = next(block_iterator)
        # Consume line with DATM
        _ = next(block_iterator)

        current_reaction = None
        for lines in block_iterator:
            if lines[0].startswith("$RFMT"):
                # A new reaction started.
                # We yield the current one before initializing the new one
                if current_reaction is not None:
                    yield current_reaction.to_reaction()
                current_reaction = ParsedReaction()

            if current_reaction is None:
                raise RuntimeError("No reaction block started")

            current_reaction.handle_line_block(lines)

        # yield the last reaction
        if current_reaction is not None:
            yield current_reaction.to_reaction()

    def iter_blocks(self) -> Iterator[List[str]]:
        current_line_block: List[str] = []
        with open(self.filename, "rt", encoding=self.encoding) as f:
            for line in f:
                line = line.rstrip("\n")

                if line.startswith("$"):
                    if current_line_block:
                        yield current_line_block
                        current_line_block = []

                current_line_block.append(line)

        # Last line block at the end of the file
        if current_line_block:
            yield current_line_block


def iterate_reactions_from_file(
    filename: Union[Path, str],
    filter_fn: Optional[Callable[[RdfReaction], bool]] = None,
) -> Iterator[RdfReaction]:
    parser = RdfParser(filename)
    reactions = (entry for entry in parser.iter_reactions())

    if filter_fn is not None:
        reactions = (reaction for reaction in reactions if filter_fn(reaction))

    yield from reactions



from rxn.chemutils.extended_reaction_smiles import (
    determine_fragment_groups,
    merge_molecules_from_fragment_groups,
    parse_extended_reaction_smiles,
    to_extended_reaction_smiles,
)
from rxn.chemutils.reaction_equation import ReactionEquation


def test_from_reaction_smiles() -> None:
    reaction_smiles = (
        "[CH3:7][C:4]1=[CH:3][C:2](=[C:1]([CH:6]=[CH:5]1)Br)[F:8]"
        ">[Li]CCCC.C1CCOC1"
        ">[CH3:7][C:4]1=[CH:3][C:2](=[C:1]([CH:6]=[CH:5]1)C(=O)O)[F:8]"
    )

    reaction = parse_extended_reaction_smiles(reaction_smiles)

    assert reaction.reactants == ["CC1=CC(F)=C(Br)C=C1"]
    assert reaction.agents == ["[Li]CCCC", "C1CCOC1"]
    assert reaction.products == ["CC1=CC(F)=C(C(=O)O)C=C1"]


def test_from_reaction_smiles_keep_atom_maps() -> None:
    reaction_smiles = (
        "[CH3:7][C:4]1=[CH:3][C:2](=[C:1]([CH:6]=[CH:5]1)Br)[F:8]"
        ">[Li]CCCC.C1CCOC1"
        ">[CH3:7][C:4]1=[CH:3][C:2](=[C:1]([CH:6]=[CH:5]1)C(=O)O)[F:8]"
    )

    reaction = parse_extended_reaction_smiles(reaction_smiles, remove_atom_maps=False)

    assert reaction.reactants == [
        "[CH3:7][C:4]1=[CH:3][C:2](=[C:1]([CH:6]=[CH:5]1)Br)[F:8]"
    ]
    assert reaction.agents == ["[Li]CCCC", "C1CCOC1"]
    assert reaction.products == [
        "[CH3:7][C:4]1=[CH:3][C:2](=[C:1]([CH:6]=[CH:5]1)C(=O)O)[F:8]"
    ]


def test_from_reaction_smiles_with_fragments() -> None:
    reaction_smiles = (
        "BrC1=C(C(=O)O)C=C(C=C1)COC.C1(O)=CC(O)=CC=C1.[OH-].[Na+]"
        ">S(=O)(=O)([O-])[O-].[Cu+2].O"
        ">OC1=CC=C2C3=C(C(OC2=C1)=O)C=C(C=C3)COC"
        " |f:2.3,4.5|"
    )

    reaction = parse_extended_reaction_smiles(reaction_smiles)

    assert reaction.reactants == [
        "BrC1=C(C(=O)O)C=C(COC)C=C1",
        "C1(O)=CC(O)=CC=C1",
        "[OH-].[Na+]",
    ]
    assert reaction.agents == ["O", "S(=O)(=O)([O-])[O-].[Cu+2]"]
    assert reaction.products == ["OC1=CC=C2C3=C(C(=O)OC2=C1)C=C(COC)C=C3"]


def test_from_reaction_smiles_with_dative_bond() -> None:
    reaction_smiles = "COC(=O)CCBr.O=C[O-]->[K+].[OH-].[Na+]>>COC(=O)CCOC(=O)C |f:2.3|"

    reaction = parse_extended_reaction_smiles(reaction_smiles)

    assert reaction.reactants == [
        "COC(=O)CCBr",
        "O=C[O-]->[K+]",
        "[OH-].[Na+]",
    ]
    assert reaction.agents == []
    assert reaction.products == ["COC(=O)CCOC(=O)C"]


def test_from_reaction_smiles_with_other_extended_information() -> None:
    # the part with "&1:", "c:", "H:", must not be considered for determination of fragments
    reaction_smiles = (
        "[CH3:1][C:2](=[O:3])[NH:4][C@H:5]([C:6](=[O:7])[OH:8])[C:9]([CH3:10])"
        "([CH3:11])[SH:12].[O:13]=[N:14]O[Na]>O.Cl.CO>[CH3:1][C:2](=[O:3])[NH:4]"
        "[C@H:5]([C:6](=[O:7])[OH:8])[C:9]([CH3:10])([CH3:11])[S:12][N:14]=[O:13]"
        " |c:5,7,H:3.2,&1:4,24,f:3.4|"
    )

    reaction = parse_extended_reaction_smiles(reaction_smiles)

    assert reaction.reactants == ["CC(=O)N[C@H](C(=O)O)C(C)(C)S", "O=NO[Na]"]
    assert reaction.agents == ["O", "Cl.CO"]
    assert reaction.products == ["CC(=O)N[C@H](C(=O)O)C(C)(C)SN=O"]


def test_from_reaction_smiles_does_not_sanitize() -> None:
    # In earlier code, molecules were canonicalized if they were part of a fragment
    reaction_smiles = "C1=CC=CC=C1.[N+](=O)(O)[O-]>>C1=CC=CC=C1N(=O)=O |f:0.1|"

    reaction = parse_extended_reaction_smiles(reaction_smiles)

    assert reaction.reactants == ["C1=CC=CC=C1.[N+](=O)(O)[O-]"]
    assert reaction.agents == []
    assert reaction.products == ["C1=CC=CC=C1N(=O)=O"]


def test_from_reaction_smiles_with_NaH() -> None:
    """
    In the original implementation using RDKit's ChemicalReaction, some molecules were parsed incorrectly, and this
    test would not have passed
    """
    reaction_smiles = (
        "C1=CC=C(C=C1)P(C2=CC=CC=C2)C3=CC=CC=C3.N(=NC(=O)OCC)C(=O)OCC.C(CO)(=O)OCC.O=C1C(CCCC1)C#N.[H-].[Na+]"
        ">C1CCOC1.C1CCOC1.C1CCOC1.C1CCOC1"
        ">C(C)OC(=O)C=1OC2=C(C1N)CCCC2"
        " |f:4.5|"
    )

    reaction = parse_extended_reaction_smiles(reaction_smiles)

    assert reaction.reactants == [
        "C1=CC=C(P(C2=CC=CC=C2)C2=CC=CC=C2)C=C1",
        "N(=NC(=O)OCC)C(=O)OCC",
        "C(CO)(=O)OCC",
        "O=C1C(C#N)CCCC1",
        "[H-].[Na+]",
    ]
    assert reaction.agents == ["C1CCOC1", "C1CCOC1", "C1CCOC1", "C1CCOC1"]
    assert reaction.products == ["C(C)OC(=O)C1=C(N)C2=C(O1)CCCC2"]


def test_from_reaction_smiles_with_rdkit_problem() -> None:
    # A previous version relying more on RDKit returned, as a reactant, 'C.[Pd]'.
    reaction_smiles = "[C].[Pd]>CC>CCO |f:0.1|"
    reaction = parse_extended_reaction_smiles(reaction_smiles)
    assert reaction.reactants == ["[C].[Pd]"]
    assert reaction.agents == ["CC"]
    assert reaction.products == ["CCO"]

    # Related example from Pistachio
    reaction_smiles = (
        "[CH3:24][C:23]([CH3:25])([CH3:26])[O:22][C:20](=[O:21])[N:10]1[C@@H:11]([CH2:12][CH2:13]"
        "[C@@H:9]1[C:4]2=[CH:3][C:2](=[C:7]([CH:6]=[CH:5]2)[F:8])[F:1])/[CH:14]=[CH:15]/[C:16]"
        "(=[O:17])[O:18][CH3:19]>[HH].[C].CCOC(=O)C.[Pd]>[CH3:24][C:23]([CH3:25])([CH3:26])[O:22]"
        "[C:20](=[O:21])[N:10]1[C@@H:11]([CH2:12][CH2:13][C@@H:9]1[C:4]2=[CH:3][C:2](=[C:7]([CH:6]"
        "=[CH:5]2)[F:8])[F:1])[CH2:14][CH2:15][C:16](=[O:17])[O:18][CH3:19] |f:2.4|"
    )
    reaction = parse_extended_reaction_smiles(reaction_smiles)
    assert reaction.reactants == [
        "CC(C)(C)OC(=O)N1[C@H](/C=C/C(=O)OC)CC[C@@H]1C1=CC(F)=C(F)C=C1"
    ]
    assert reaction.agents == ["[HH]", "CCOC(=O)C", "[C].[Pd]"]
    assert reaction.products == [
        "CC(C)(C)OC(=O)N1[C@H](CCC(=O)OC)CC[C@@H]1C1=CC(F)=C(F)C=C1"
    ]


def test_to_reaction_smiles() -> None:
    reactants = ["CC1=CC(F)=C(Br)C=C1"]
    agents = ["[Li]CCCC", "C1CCOC1"]
    products = ["CC1=CC(F)=C(C(=O)O)C=C1"]

    reaction = ReactionEquation(reactants, agents, products)

    reaction_smiles = to_extended_reaction_smiles(reaction)
    expected = "CC1=CC(F)=C(Br)C=C1" ">[Li]CCCC.C1CCOC1" ">CC1=CC(F)=C(C(=O)O)C=C1"

    assert reaction_smiles == expected


def test_to_reaction_smiles_with_fragments() -> None:
    reactants = ["BrC1=C(C(=O)O)C=C(COC)C=C1", "C1(O)=CC(O)=CC=C1", "[Na+].[OH-]"]
    agents = ["O", "O=S(=O)([O-])[O-].[Cu+2]"]
    products = ["OC1=CC=C2C3=C(C(=O)OC2=C1)C=C(COC)C=C3"]

    reaction = ReactionEquation(reactants, agents, products)

    reaction_smiles = to_extended_reaction_smiles(reaction)
    expected = (
        "BrC1=C(C(=O)O)C=C(COC)C=C1.C1(O)=CC(O)=CC=C1.[Na+].[OH-]"
        ">O.O=S(=O)([O-])[O-].[Cu+2]"
        ">OC1=CC=C2C3=C(C(=O)OC2=C1)C=C(COC)C=C3"
        " |f:2.3,5.6|"
    )

    assert reaction_smiles == expected


def test_to_reaction_smiles_with_dummy_fragments() -> None:
    reactants = ["A", "B.C", "D", "E"]
    agents = ["F.G", "H"]
    products = ["I", "J", "K", "L.M.N"]

    reaction = ReactionEquation(reactants, agents, products)

    reaction_smiles = to_extended_reaction_smiles(reaction)
    expected = "A.B.C.D.E" ">F.G.H" ">I.J.K.L.M.N" " |f:1.2,5.6,11.12.13|"

    assert reaction_smiles == expected


def test_fragment_groups() -> None:
    assert determine_fragment_groups("") == []

    # Cases with only f involved
    assert determine_fragment_groups("|f:2.3,4.5|") == [[2, 3], [4, 5]]
    assert determine_fragment_groups("|f:2.3,4.5.8|") == [[2, 3], [4, 5, 8]]
    assert determine_fragment_groups("|f:2.3,4|") == [[2, 3], [4]]
    assert determine_fragment_groups("|f:12.3,4.105|") == [[12, 3], [4, 105]]

    # More complex cases with other information; may contain more information, see
    # https://docs.chemaxon.com/display/docs/chemaxon-extended-smiles-and-smarts-cxsmiles-and-cxsmarts.md
    assert determine_fragment_groups("|a1:3,4,f:2.3,4|") == [[2, 3], [4]]
    assert determine_fragment_groups("|f:2.3,4,&2:3,55,6|") == [[2, 3], [4]]
    assert determine_fragment_groups("|m:0,22,f:2.3,4,&2:3,55,6|") == [[2, 3], [4]]


def test_merge_molecules_from_fragment_groups() -> None:
    # We consider, as an example, the following reaction
    # C.CC.CCC.CCCC > N.NN.NNN.NNNN.NNNNN > O.OO.OOO.OOOO
    # with fragmentation info |f:0.2.3,4.6,5.8|

    groups = [[0, 2, 3], [4, 6], [5, 8]]

    assert merge_molecules_from_fragment_groups(
        ["C", "CC", "CCC", "CCCC"], groups, 0
    ) == ["CC", "C.CCC.CCCC"]
    assert merge_molecules_from_fragment_groups(
        ["N", "NN", "NNN", "NNNN", "NNNNN"], groups, 4
    ) == ["NNNN", "N.NNN", "NN.NNNNN"]
    assert merge_molecules_from_fragment_groups(
        ["O", "OO", "OOO", "OOOO"], groups, 9
    ) == ["O", "OO", "OOO", "OOOO"]



from typing import Optional

from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdmolfiles import MolToSmiles


class InvalidSmiles(ValueError):
    def __init__(self, smiles: str, msg: Optional[str] = None):
        if msg is None:
            msg = f'"{smiles}" is not a valid SMILES string'
        super().__init__(msg)
        self.smiles = smiles


class InvalidInchi(ValueError):
    """
    Exception raised when converting invalid InChI strings.
    """

    def __init__(self, inchi: str):
        super().__init__(f"The following InChI string cannot be converted: {inchi}")
        self.inchi = inchi


class InvalidReactionSmiles(InvalidSmiles):
    def __init__(self, reaction_smiles: str, msg: Optional[str] = None):
        if msg is None:
            msg = f'"{reaction_smiles}" is not a valid reaction SMILES string'
        super().__init__(smiles=reaction_smiles, msg=msg)


class InvalidMdl(ValueError):
    """
    Exception raised when converting invalid MDL Mol strings.
    """

    def __init__(self, mdl: str):
        super().__init__(f"The following MDL string cannot be converted: {mdl}")
        self.mdl = mdl


class SanitizationError(ValueError):
    def __init__(self, mol: Mol):
        message = "Error when sanitizing RDKit Mol"
        try:
            smiles = MolToSmiles(mol)
            message += ": " + smiles
        except Exception:
            pass
        super().__init__(message)


class UnclearWhetherTokenized(ValueError):
    """Exception raised when unclear if something was tokenized or not."""

    def __init__(self, string: str):
        super().__init__(f'Cannot determine if "{string}" is tokenized.')



import pytest

from rxn.chemutils.reaction_combiner import ReactionCombiner
from rxn.chemutils.reaction_smiles import ReactionFormat


def test_reaction_combiner_on_precursors_and_products() -> None:
    combiner = ReactionCombiner()

    precursors = ["CC.O", "CCC.O"]
    products = ["CCO", "CCCO"]
    expected = ["CC.O>>CCO", "CCC.O>>CCCO"]

    assert list(combiner.combine(precursors, products)) == expected


def test_reaction_combiner_on_fragment_reactions() -> None:
    combiner = ReactionCombiner()

    fragment_1 = ["CC.O>>", "CCC>>CCCO"]
    fragment_2 = [">>CCO", "O.N>>"]
    expected = ["CC.O>>CCO", "CCC.O.N>>CCCO"]

    assert list(combiner.combine(fragment_1, fragment_2)) == expected


def test_reaction_combiner_with_tokenized_input() -> None:
    combiner = ReactionCombiner()

    fragment_1 = ["C C . O", "C C C >> C C C O"]
    fragment_2 = ["C C O", "O . N >>"]
    expected = ["CC.O>>CCO", "CCC.O.N>>CCCO"]

    assert list(combiner.combine(fragment_1, fragment_2)) == expected


def test_multiple_precursors_per_product() -> None:
    combiner = ReactionCombiner()

    # Three sets of precursors for one product
    # Note: the same would work for two sets of fragment reactions.
    precursors = ["CC.O", "CC.O.N", "CC.O.P", "CCC.O", "CCC.O.N", "CCC.O.P"]
    products = ["CCO", "CCCO"]
    expected = [
        "CC.O>>CCO",
        "CC.O.N>>CCO",
        "CC.O.P>>CCO",
        "CCC.O>>CCCO",
        "CCC.O.N>>CCCO",
        "CCC.O.P>>CCCO",
    ]

    assert list(combiner.combine(precursors, products)) == expected


def test_multiple_products_per_precursors() -> None:
    combiner = ReactionCombiner()

    # Two products for each set of precursors
    # Note: the same would work for two sets of fragment reactions.
    precursors = ["CC.O", "CCC.O"]
    products = ["CCO", "OCCO", "CCCO", "OCCCO"]
    expected = [
        "CC.O>>CCO",
        "CC.O>>OCCO",
        "CCC.O>>CCCO",
        "CCC.O>>OCCCO",
    ]

    assert list(combiner.combine(precursors, products)) == expected


def test_incompatible_number_of_precursors_and_products() -> None:
    combiner = ReactionCombiner()

    # Three precursors and two products -> unclear how to combine
    precursors = ["CC.O", "CCC.O", "CCCC.O"]
    products = ["CCO", "CCCO"]

    with pytest.raises(ValueError):
        _ = list(combiner.combine(precursors, products))


def test_different_output_formats() -> None:
    fragment_1 = ["CC~O", "CCC>>CCCO"]
    fragment_2 = ["CCO", "N~O>>"]

    assert list(
        ReactionCombiner(reaction_format=ReactionFormat.STANDARD).combine(
            fragment_1, fragment_2
        )
    ) == ["CC.O>>CCO", "CCC.N.O>>CCCO"]
    assert list(
        ReactionCombiner(reaction_format=ReactionFormat.STANDARD_WITH_TILDE).combine(
            fragment_1, fragment_2
        )
    ) == ["CC~O>>CCO", "CCC.N~O>>CCCO"]
    assert list(
        ReactionCombiner(reaction_format=ReactionFormat.EXTENDED).combine(
            fragment_1, fragment_2
        )
    ) == ["CC.O>>CCO |f:0.1|", "CCC.N.O>>CCCO |f:1.2|"]


def test_standardization() -> None:
    precursors = ["C(C).O", "O.C(CC)"]
    products = ["CCO", "CCCO"]

    assert list(ReactionCombiner(standardize=True).combine(precursors, products)) == [
        "CC.O>>CCO",
        "CCC.O>>CCCO",
    ]
    assert list(ReactionCombiner(standardize=False).combine(precursors, products)) == [
        "C(C).O>>CCO",
        "O.C(CC)>>CCCO",
    ]


def test_invalid_reactions() -> None:
    fallback_string = "C>>C"
    combiner = ReactionCombiner(standardize=True, fallback_reaction=fallback_string)

    precursors = ["CC.O", "CCC.O", "CC>CC>O"]
    products = ["CCo", "CCCO", "CCCCO"]
    expected = [fallback_string, "CCC.O>>CCCO", fallback_string]

    assert list(combiner.combine(precursors, products)) == expected


def test_combine_iterators() -> None:
    combiner = ReactionCombiner()

    precursors = iter(["CC.O", "CCC.O"])
    products = iter(["CCO", "CCCO"])
    expected = ["CC.O>>CCO", "CCC.O>>CCCO"]

    assert list(combiner.combine_iterators(precursors, products)) == expected


def test_combine_iterators_with_multiplier() -> None:
    combiner = ReactionCombiner()

    precursors = iter(["CC.O", "CC.O.N", "CCC.O", "CCC.O.N"])
    products = iter(["CCO", "CCCO"])
    expected = ["CC.O>>CCO", "CC.O.N>>CCO", "CCC.O>>CCCO", "CCC.O.N>>CCCO"]

    results = combiner.combine_iterators(precursors, products, 1, 2)
    assert list(results) == expected


def test_raise_if_iterator_not_fully_consumed() -> None:
    combiner = ReactionCombiner()

    # One more precursor than supposed to be ("CC")
    precursors = iter(["CC.O", "CC.O.N", "CCC.O", "CCC.O.N", "CC"])
    products = iter(["CCO", "CCCO"])
    with pytest.raises(RuntimeError):
        _ = list(combiner.combine_iterators(precursors, products, 1, 2))

    # One more product than supposed to be ("CC")
    precursors = iter(["CC.O", "CC.O.N", "CCC.O", "CCC.O.N"])
    products = iter(["CCO", "CCCO", "CC"])
    with pytest.raises(RuntimeError):
        _ = list(combiner.combine_iterators(precursors, products, 1, 2))



import sys
from typing import Optional, TextIO

import click
from rxn.utilities.logging import setup_console_logger

from rxn.chemutils.tokenization import tokenize_smiles


@click.command()
@click.argument("input_file", type=click.File(mode="r"), default=sys.stdin)
@click.argument("output_file", type=click.File(mode="w"), default=sys.stdout)
@click.option(
    "--fallback_value",
    type=str,
    help="Placeholder for strings that cannot be tokenized. By default, an exception is raised.",
)
def main(
    input_file: TextIO, output_file: TextIO, fallback_value: Optional[str]
) -> None:
    """
    Tokenize SMILES strings (molecules or reactions).

    The script will read strings either from stdin, or from a file given as the
    first argument, and write to stdout, or from a file given as the second
    argument.
    """
    setup_console_logger()

    for line in input_file:
        smiles = line.strip()
        output_file.write(f"{tokenize_smiles(smiles, fallback_value=fallback_value)}\n")


if __name__ == "__main__":
    main()



import sys
from typing import Optional, TextIO

import click
from rxn.utilities.logging import setup_console_logger

from rxn.chemutils.miscellaneous import canonicalize_any


@click.command()
@click.argument("input_file", type=click.File(mode="r"), default=sys.stdin)
@click.argument("output_file", type=click.File(mode="w"), default=sys.stdout)
@click.option(
    "--invalid_placeholder",
    type=str,
    help=(
        "If specified, the given value will be the output for invalid input SMILES. "
        "By default, an exception is raised in such cases."
    ),
)
@click.option(
    "--sort_compounds",
    "-s",
    is_flag=True,
    help="If specified, the compounds will be sorted after canonicalization.",
)
def main(
    input_file: TextIO,
    output_file: TextIO,
    invalid_placeholder: Optional[str],
    sort_compounds: bool,
) -> None:
    """
    Canonicalize SMILES strings (molecules, sets of molecules, or reactions).

    The script will read strings either from stdin, or from a file given as the
    first argument, and write to stdout, or from a file given as the second
    argument.
    """
    setup_console_logger()

    for line in input_file:
        smiles = line.strip()

        # Canonicalize the SMILES, handle exception if needed
        try:
            canonical = canonicalize_any(smiles, sort_molecules=sort_compounds)
        except Exception:
            if invalid_placeholder is None:
                raise
            canonical = invalid_placeholder

        output_file.write(f"{canonical}\n")


if __name__ == "__main__":
    main()



from .rdf_parser import RdfParser, iterate_reactions_from_file
from .rdf_reaction import RdfReaction
from .reaction_smiles_extractor import ReactionSmilesExtractor

__all__ = [
    "RdfParser",
    "RdfReaction",
    "ReactionSmilesExtractor",
    "iterate_reactions_from_file",
]



from typing import List

import pytest

from rxn.chemutils.conversion import canonicalize_smiles
from rxn.chemutils.exceptions import InvalidReactionSmiles, InvalidSmiles
from rxn.chemutils.reaction_equation import (
    ReactionEquation,
    apply_to_compound_groups,
    apply_to_compounds,
    canonicalize_compounds,
    cleanup_compounds,
    has_repeated_molecules,
    merge_reactants_and_agents,
    remove_duplicate_compounds,
    remove_precursors_from_products,
    rxn_standardization,
    sort_compounds,
)


def test_merge_reactants_and_agents() -> None:
    reactants = ["COCO", "CCOC", "OCC"]
    agents = ["O", "C", "N"]
    products = ["NCOC", "CCNO", "NCCC"]

    reaction = ReactionEquation(reactants, agents, products)

    merged_reaction = merge_reactants_and_agents(reaction)

    expected = ReactionEquation(reactants + agents, [], products)
    assert merged_reaction == expected


def test_sort_compounds() -> None:
    reactants = ["COCO", "CCOC", "OCC"]
    agents = ["O", "C", "N"]
    products = ["NCOC", "CCNO", "NCCC"]

    reaction = ReactionEquation(reactants, agents, products)

    sorted_reaction = sort_compounds(reaction)

    expected = ReactionEquation(sorted(reactants), sorted(agents), sorted(products))
    assert sorted_reaction == expected


def test_canonicalize_compounds() -> None:
    reactants = ["COCO", "CCOC", "OCC"]
    agents = ["O", "C", "N"]
    products = ["NCOC", "CCNO", "NCCC"]

    reaction = ReactionEquation(reactants, agents, products)

    canonicalized_reaction = canonicalize_compounds(reaction)

    expected = ReactionEquation(
        [canonicalize_smiles(s) for s in reactants],
        [canonicalize_smiles(s) for s in agents],
        [canonicalize_smiles(s) for s in products],
    )
    assert canonicalized_reaction == expected


def test_canonicalize_compounds_with_invalid_valence() -> None:
    reaction = ReactionEquation(reactants=["F", "CC"], agents=[], products=["CFC"])

    with pytest.raises(InvalidSmiles):
        canonicalize_compounds(reaction)

    # does not raise if no valence check, returns original reaction
    assert canonicalize_compounds(reaction, check_valence=False) == reaction


def test_cleanup_compounds() -> None:
    reaction = ReactionEquation(["CC[CH2]C", "C-C"], [], ["[C].[Pd]"])

    clean_reaction = cleanup_compounds(reaction)
    expected = ReactionEquation(["CCCC", "CC"], [], ["[C].[Pd]"])

    assert clean_reaction == expected


def test_remove_duplicate_compounds() -> None:
    reaction = ReactionEquation(
        ["A", "B", "C", "A", "B"], ["D", "D", "E"], ["A", "B", "G", "H", "H"]
    )

    reaction_without_duplicates = remove_duplicate_compounds(reaction)
    # The compounds are in the same order, and duplicates in different categories are not removed
    expected = ReactionEquation(["A", "B", "C"], ["D", "E"], ["A", "B", "G", "H"])

    assert reaction_without_duplicates == expected


def test_equation_to_string() -> None:
    reactants = ["COCO", "[Na+].[OH-]", "OCC"]
    agents = ["O", "C"]
    products = ["NCOC"]
    reaction = ReactionEquation(reactants, agents, products)

    reaction_string = reaction.to_string()
    expected = "COCO.[Na+].[OH-].OCC>O.C>NCOC"

    assert reaction_string == expected


def test_equation_to_string_with_fragment_bond() -> None:
    reactants = ["COCO", "[Na+].[OH-]", "OCC"]
    agents = ["O", "C"]
    products = ["NCOC"]
    reaction = ReactionEquation(reactants, agents, products)

    reaction_string = reaction.to_string(fragment_bond="~")
    expected = "COCO.[Na+]~[OH-].OCC>O.C>NCOC"

    assert reaction_string == expected


def test_equation_from_string() -> None:
    reaction_string = "COCO.[Na+].[OH-].OCC>O.C>NCOC"

    reaction = ReactionEquation.from_string(reaction_string)

    expected_reactants = ["COCO", "[Na+]", "[OH-]", "OCC"]
    expected_agents = ["O", "C"]
    expected_products = ["NCOC"]
    expected_reaction = ReactionEquation(
        expected_reactants, expected_agents, expected_products
    )

    assert reaction == expected_reaction


def test_equation_from_string_ignores_additional_dots() -> None:
    reaction_smiles = "..A.B.>.>C.D."
    expected = ReactionEquation(["A", "B"], [], ["C", "D"])
    assert ReactionEquation.from_string(reaction_smiles) == expected


def test_equation_from_string_with_invalid_strings() -> None:
    invalid_reaction_smiles = ["A.B>C", "A>>>C", "A>>B>>C", "A.B.C"]
    for reaction_smiles in invalid_reaction_smiles:
        with pytest.raises(InvalidReactionSmiles):
            _ = ReactionEquation.from_string(reaction_smiles)


def test_equation_from_string_with_fragment_bond() -> None:
    reaction_string = "COCO.[Na+]~[OH-].OCC>O.C>NCOC"

    reaction = ReactionEquation.from_string(reaction_string, fragment_bond="~")

    expected_reactants = ["COCO", "[Na+].[OH-]", "OCC"]
    expected_agents = ["O", "C"]
    expected_products = ["NCOC"]
    expected_reaction = ReactionEquation(
        expected_reactants, expected_agents, expected_products
    )

    assert reaction == expected_reaction


def test_equation_from_string_with_no_agent() -> None:
    reaction_string = "COCO.[Na+].[OH-].OCC>>NCOC"

    reaction = ReactionEquation.from_string(reaction_string)

    expected_reactants = ["COCO", "[Na+]", "[OH-]", "OCC"]
    expected_agents: List[str] = []
    expected_products = ["NCOC"]
    expected_reaction = ReactionEquation(
        expected_reactants, expected_agents, expected_products
    )

    assert reaction == expected_reaction


def test_equation_from_string_with_dative_bond() -> None:
    reaction_string = "COC(=O)CCBr.O=C([O-]->[K+])>>COC(=O)CCOC(=O)C"

    reaction = ReactionEquation.from_string(reaction_string)

    expected_reactants = ["COC(=O)CCBr", "O=C([O-]->[K+])"]
    expected_agents: List[str] = []
    expected_products = ["COC(=O)CCOC(=O)C"]
    expected_reaction = ReactionEquation(
        expected_reactants, expected_agents, expected_products
    )

    assert reaction == expected_reaction


def test_iter_all_smiles() -> None:
    reaction_string = "COCO.[Na+].[OH-]>O>NCOC"
    reaction = ReactionEquation.from_string(reaction_string)
    assert list(reaction.iter_all_smiles()) == ["COCO", "[Na+]", "[OH-]", "O", "NCOC"]


def test_has_repeated_molecules() -> None:
    assert not has_repeated_molecules(ReactionEquation(["A", "B"], ["C"], ["D", "E"]))

    assert has_repeated_molecules(ReactionEquation(["A", "A"], ["C"], ["D", "E"]))
    assert has_repeated_molecules(ReactionEquation(["A", "B"], ["C", "C"], ["D", "E"]))
    assert has_repeated_molecules(ReactionEquation(["A", "B"], ["C"], ["D", "E", "E"]))
    assert has_repeated_molecules(ReactionEquation(["A", "B"], ["A"], ["D", "E"]))
    assert has_repeated_molecules(ReactionEquation(["A", "B"], ["C"], ["B", "E"]))


def test_apply_to_compounds() -> None:
    def dummy_fn(smiles: str) -> str:
        # add a 'C' to the SMILES
        return smiles + "C"

    assert apply_to_compounds(
        ReactionEquation(["C", "N", "O"], ["P"], ["CNO"]), dummy_fn
    ) == ReactionEquation(["CC", "NC", "OC"], ["PC"], ["CNOC"])


def test_apply_to_compound_groups() -> None:
    def dummy_fn(compound_group: List[str]) -> List[str]:
        # as a test: reverse the order of the compounds
        return list(reversed(compound_group))

    assert apply_to_compound_groups(
        ReactionEquation(["C", "N", "O"], ["P"], ["CNO", "C"]), dummy_fn
    ) == ReactionEquation(["O", "N", "C"], ["P"], ["C", "CNO"])


def test_rxn_standardization() -> None:
    initial_reaction_equation = ReactionEquation(
        reactants=["OCC", "O", "C"],
        agents=["CCO"],
        products=["C"],
    )

    # NB:
    # 1. CCO is merged into the reactants
    # 2. OCC is canonicalized to CCO
    # 3. CCO is now present two times and removed
    # 4. sorted alphabetically
    expected = ReactionEquation(
        reactants=["C", "CCO", "O"],
        agents=[],
        products=["C"],
    )

    assert rxn_standardization(initial_reaction_equation) == expected


def test_remove_precursors_from_products() -> None:
    reactants = ["COCO", "[Na+].[OH-]", "OCC"]
    agents = ["O", "C"]
    products = ["NCOC", "C", "COCO"]
    reaction = ReactionEquation(reactants, agents, products)

    reaction = remove_precursors_from_products(reaction)

    expected_products = ["NCOC"]
    expected_reaction = ReactionEquation(reactants, agents, expected_products)

    assert reaction == expected_reaction


def test_can_be_instantiated_from_any_iterator() -> None:
    a = {"C", "O"}
    b = (m for m in ["CO", "OC"])
    reaction_equation = ReactionEquation(a, [], b)

    assert reaction_equation.reactants == ["C", "O"] or reaction_equation.reactants == [
        "O",
        "C",
    ]
    assert reaction_equation.products == ["CO", "OC"]


def test_list_objects_are_not_shared() -> None:
    a = ["C", "O"]
    b = ["CO"]
    reaction_equation = ReactionEquation(a, [], b)

    # Modifying a list after using it to instantiate ReactionEquation should
    # have no impact on the ReactionEquation instance
    assert reaction_equation.reactants == a
    a.append("N")
    assert reaction_equation.reactants != a

    # Modifying the ReactionEquation compounds lists should have no impact on
    # the lists that were used to instantiate it.
    assert b == reaction_equation.products
    reaction_equation.products.append("N")
    assert b != reaction_equation.products



from typing import Iterable, List, Optional

from rxn.chemutils.reaction_equation import ReactionEquation

from ..conversion import mdl_to_smiles
from .rdf_reaction import RdfReaction
from .reaction_properties import ReactionProperties


class ReactionSmilesExtractor:
    """Extract reaction SMILES from RdfReaction instances."""

    def __init__(self, fragment_bond: Optional[str] = None, sanitize: bool = True):
        self.fragment_bond = fragment_bond
        self.sanitize = sanitize

    def to_reaction_smiles(self, reaction: RdfReaction) -> str:
        """Extract the reaction SMILES.

        Args:
            reaction: RdfReaction to extract the reaction equation from.
        """
        return self.to_string(self.to_reaction_equation(reaction))

    def to_reaction_equation(self, reaction: RdfReaction) -> ReactionEquation:
        """Extract the reaction equation.

        Args:
            reaction: RdfReaction to extract the reaction equation from.
        """

        # NB: reaction.reagents should actually always be an empty list
        reagents = reaction.reagents.copy()
        reagents.extend(
            c.mol_structure for c in ReactionProperties(reaction.meta).get_compounds()
        )

        return ReactionEquation(
            reactants=self._to_smiles_group(reaction.reactants),
            agents=self._to_smiles_group(reagents),
            products=self._to_smiles_group(reaction.products),
        )

    def to_string(self, reaction_equation: ReactionEquation) -> str:
        """Convert a ReactionEquation to a reaction SMILES.

        Useful function as it automatically uses the correct fragment bond."""
        return reaction_equation.to_string(fragment_bond=self.fragment_bond)

    def _to_smiles(self, mdl: str) -> str:
        return mdl_to_smiles(mdl, sanitize=self.sanitize, canonicalize=True)

    def _to_smiles_group(self, mdl_iterable: Iterable[str]) -> List[str]:
        return [self._to_smiles(m) for m in mdl_iterable]



import logging
import random
from typing import Callable, List

from .miscellaneous import apply_to_any_smiles, apply_to_smiles_groups

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class SmilesAugmenter:
    """
    Class to augment any kind of SMILES string with the help of randomization
    and shuffling.
    """

    def __init__(
        self,
        augmentation_fn: Callable[[str], str],
        augmentation_probability: float = 1.0,
        shuffle: bool = True,
        ignore_exceptions: bool = True,
    ):
        """
        Args:
            augmentation_fn: Function for augmenting the individual SMILES strings,
                such as the functions provided in smiles_randomization.py.
            augmentation_probability: Probability with which to augment individual
                SMILES strings.
            shuffle: Whether to shuffle the order of the compounds.
            ignore_exceptions: Whether to ignore the error (and return the
                original string) when an augmentation fails. If False, exceptions
                will be propagated.
        """
        self.augmentation_fn = augmentation_fn
        self.augmentation_probability = augmentation_probability
        self.shuffle = shuffle
        self.ignore_exceptions = ignore_exceptions

    def augment(self, smiles: str, number_augmentations: int) -> List[str]:
        """
        Augment one SMILES string (of any kind).

        Args:
            smiles: SMILES string to augment.
            number_augmentations: how many times to do the augmentation.
        """

        # augmentation of the individual compound SMILES
        augmented = [
            apply_to_any_smiles(
                smiles, self._augment_with_probability, force_multicomponent=True
            )
            for _ in range(number_augmentations)
        ]

        # shuffle the order of the compounds
        if self.shuffle:
            augmented = [
                apply_to_smiles_groups(s, SmilesAugmenter._shuffle) for s in augmented
            ]

        return augmented

    def _augment_with_probability(self, smiles: str) -> str:
        """Augmentat a SMILES, with the probability given by the member variable."""

        # Note: no need to call random.uniform if the augmentation probability is 1.0.
        if (
            self.augmentation_probability == 1.0
            or random.uniform(0, 1) <= self.augmentation_probability
        ):
            try:
                return self.augmentation_fn(smiles)
            except Exception as e:
                if self.ignore_exceptions:
                    logger.warning(f"Augmentation failed for {smiles}: {e}")
                    return smiles
                else:
                    raise

        # no augmentation
        return smiles

    @staticmethod
    def _shuffle(smiles_list: List[str]) -> List[str]:
        smiles_list = smiles_list.copy()
        random.shuffle(smiles_list)
        return smiles_list



import logging
import re
import typing
from collections import Counter
from functools import partial
from typing import Callable, List

from rdkit.Chem import AddHs, Atom, Mol
from rxn.utilities.files import (
    PathLike,
    dump_list_to_file,
    iterate_lines_from_file,
    raise_if_paths_are_identical,
)

from .conversion import canonicalize_smiles, smiles_to_mol
from .exceptions import InvalidSmiles
from .multicomponent_smiles import (
    apply_to_multicomponent_smiles,
    list_to_multicomponent_smiles,
    multicomponent_smiles_to_list,
    sort_multicomponent_smiles,
)
from .reaction_equation import (
    ReactionEquation,
    apply_to_compound_groups,
    apply_to_compounds,
    sort_compounds,
)
from .reaction_smiles import (
    determine_format,
    parse_any_reaction_smiles,
    parse_reaction_smiles,
    to_reaction_smiles,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

CHIRAL_CENTER_PATTERN = re.compile(
    r"\[([^],@]+)@+([^]]*)]"
)  # Matches stereo centres, and groups what comes before and after "@"


def is_valid_smiles(smiles: str, check_valence: bool = True) -> bool:
    """
    Whether a given string corresponds to a valid SMILES string.

    Args:
        smiles: string to check.
        check_valence: whether to check the valence.

    Returns:
        True if the given SMILES is valid, else False.
    """
    try:
        canonicalize_smiles(smiles, check_valence=check_valence)
        return True
    except InvalidSmiles:
        return False


def equivalent_smiles(*smiles: str, check_valence: bool = False) -> bool:
    """
    Returns true if all the given SMILES strings are equivalent.
    Will catch the exceptions for invalid SMILES and return false in that case.

    Args:
        smiles: multiple SMILES to check for equivalence.
        check_valence: if True, molecules with invalid valence will be invalidated.
    """
    try:
        canonical_smiles = [
            canonicalize_smiles(s, check_valence=check_valence) for s in smiles
        ]
        return len(set(canonical_smiles)) == 1
    except InvalidSmiles:
        return False


def atom_type_counter(smiles: str) -> typing.Counter[str]:
    """
    Return a counter of atom types (as symbols).
    """

    mol: Mol = AddHs(smiles_to_mol(smiles, sanitize=False))
    atoms: List[Atom] = mol.GetAtoms()  # type: ignore[call-arg,no-untyped-call]
    return Counter(atom.GetSymbol() for atom in atoms)


def remove_chiral_centers(smiles: str) -> str:
    """
    Return SMILES where all the chiral centres are removed.

    Args:
        smiles: non-atom-mapped SMILES string.

    Returns:
        SMILES with no chiral information. It is not canonical.
    """
    return re.sub(CHIRAL_CENTER_PATTERN, r"[\g<1>\g<2>]", smiles)


def remove_double_bond_stereochemistry(smiles: str) -> str:
    """
    Return SMILES where all the E/Z information on double bonds is removed.

    Args:
        smiles: SMILES string.

    Returns:
        SMILES with no sterochemical information for double bonds. The SMILES
        is not guaranteed to be canonical.
    """
    return smiles.replace("/", "").replace("\\", "")


def apply_to_any_smiles(
    any_smiles: str, fn: Callable[[str], str], force_multicomponent: bool = False
) -> str:
    """
    Apply a given function to individual compound SMILES strings given in any kind
    of SMILES string (molecule SMILES, multicomponent SMILES, reaction SMILES).

    In the case of reaction SMILES, the format is kept.

    Args:
        any_smiles: any kind of SMILES string.
        fn: callback to apply to every compound SMILES.
        force_multicomponent: by default, when a SMILES string contains no ">" or "~",
            it is assumed to just be a normal single-component SMILES string. Providing
            force_multicomponent=True leads to an interpretation as a multismiles string,
            i.e. splitting at all the dots.

    Raises:
        Exception: different kinds of exception may be raised during parsing,
            or during execution of the callback.

    Returns:
        the new (molecule, multicomponent, or reaction) SMILES string after
        application of the callback to all the component SMILES.
    """
    if ">" in any_smiles:
        # we have a reaction SMILES
        reaction_format = determine_format(any_smiles)
        reaction = parse_reaction_smiles(any_smiles, reaction_format)
        reaction = apply_to_compounds(reaction, fn)
        return to_reaction_smiles(reaction, reaction_format)
    elif "~" in any_smiles or force_multicomponent:
        # we have a multicomponent SMILES
        return apply_to_multicomponent_smiles(any_smiles, fn=fn, fragment_bond="~")
    else:
        # we have a single-component SMILES
        return fn(any_smiles)


def apply_to_smiles_groups(
    any_smiles: str, fn: Callable[[List[str]], List[str]]
) -> str:
    """
    Apply a given function to groups of SMILES strings given in any
    multicomponent SMILES or reaction SMILES.

    This function can typically be used for sorting or shuffling precursors,
    products, etc.

    In the case of reaction SMILES, the format is kept.

    Args:
        any_smiles: any kind of SMILES string.
        fn: callback to apply to every compound SMILES.

    Raises:
        Exception: different kinds of exception may be raised during parsing,
            or during execution of the callback.

    Returns:
        the new (multicomponent, or reaction) SMILES string after
        application of the callback to all the groups of SMILES.
    """
    if ">" in any_smiles:
        # we have a reaction SMILES
        reaction_format = determine_format(any_smiles)
        reaction = parse_reaction_smiles(any_smiles, reaction_format)
        reaction = apply_to_compound_groups(reaction, fn)
        return to_reaction_smiles(reaction, reaction_format)
    else:
        # we have a multicomponent SMILES
        compounds = multicomponent_smiles_to_list(any_smiles, fragment_bond="~")
        return list_to_multicomponent_smiles(fn(compounds), fragment_bond="~")


def canonicalize_any(
    any_smiles: str,
    check_valence: bool = True,
    sort_molecules: bool = False,
    fallback_value: typing.Optional[str] = None,
) -> str:
    """
    Canonicalize any SMILES string (molecule SMILES, multicomponent SMILES, reaction SMILES).

    In the case of reaction SMILES, the format is kept.

    Args:
        any_smiles: any kind of SMILES string.
        check_valence: if False, will not do any valence check.
        sort_molecules: whether to sort the compounds alphabetically at the same time.
        fallback_value: what value to returns when the canonicalization is unsuccessful.
            Default: no fallback, will propagate the exception.

    Raises:
        Exception: different kinds of exception may be raised during parsing.
        InvalidSmiles: for canonicalization errors.

    Returns:
        the canonical (molecule, multicomponent, or reaction) SMILES string.
    """
    try:
        fn = partial(canonicalize_smiles, check_valence=check_valence)
        canonical_smiles = apply_to_any_smiles(any_smiles, fn)
        if sort_molecules:
            canonical_smiles = sort_any(canonical_smiles)
        return canonical_smiles
    except Exception as e:
        if fallback_value is not None:
            logger.debug(f'Error when canonicalizing "{any_smiles}": {e}')
            return fallback_value
        raise


def canonicalize_file(
    input_file: PathLike,
    output_file: PathLike,
    check_valence: bool = True,
    fallback_value: str = "",
    sort_molecules: bool = False,
) -> None:
    raise_if_paths_are_identical(input_file, output_file)
    logger.info(f'Canonicalizing file "{input_file}" -> "{output_file}".')

    # We formulate it as a generator, so that the file below is written directly
    canonical = (
        canonicalize_any(
            line,
            check_valence=check_valence,
            fallback_value=fallback_value,
            sort_molecules=sort_molecules,
        )
        for line in iterate_lines_from_file(input_file)
    )

    dump_list_to_file(canonical, output_file)


def sort_any(any_smiles: str) -> str:
    """
    Sort any SMILES string (molecule SMILES, multicomponent SMILES, reaction SMILES).

    For single-component SMILES, the fragments will be reorderd.
    In the case of reaction SMILES, the format is kept.

    Args:
        any_smiles: any kind of SMILES string.

    Raises:
        Exception: different kinds of exception may be raised during parsing.

    Returns:
        the sorted SMILES string.
    """
    if ">" in any_smiles:
        # we have a reaction SMILES
        reaction_format = determine_format(any_smiles)
        reaction = parse_reaction_smiles(any_smiles, reaction_format)
        reaction = sort_compounds(reaction)
        return to_reaction_smiles(reaction, reaction_format)
    else:
        # we call the same function for single- and multi-component SMILES
        return sort_multicomponent_smiles(any_smiles)


def get_individual_compounds(any_smiles: str) -> List[str]:
    """
    Get the individual compound SMILES strings starting from any SMILES string
    (multicomponent SMILES, reaction SMILES).

    Single-component SMILES with dots are interpreted as multicomponent SMILES strings.

    Args:
        any_smiles: any kind of SMILES string.

    Raises:
        Exception: different kinds of exception may be raised during parsing.

    Returns:
        List of individual compound SMILES.
    """
    if ">" in any_smiles:
        # We have a reaction SMILES
        reaction = parse_any_reaction_smiles(any_smiles)
        return list(reaction.iter_all_smiles())
    else:
        # We interpret it as a multicomponent SMILES.
        # We use "~" as a fragment bond even if it is not actually needed.
        return multicomponent_smiles_to_list(any_smiles, fragment_bond="~")


def merge_reactions(*reactions: ReactionEquation) -> ReactionEquation:
    """Merge several reactions into one.

    Useful when ReactionEquation is used to store partial equations."""
    reactants = []
    agents = []
    products = []
    for reaction in reactions:
        reactants.extend(reaction.reactants)
        agents.extend(reaction.agents)
        products.extend(reaction.products)
    return ReactionEquation(reactants=reactants, agents=agents, products=products)


def mol_has_atom_mapping(mol: Mol) -> bool:
    """
    Whether at least one atom of an RDKit Mol contains an atom map number.

    Args:
        mol: RDKit Mol.
    """
    atom: Atom
    for atom in mol.GetAtoms():  # type: ignore[call-arg,no-untyped-call]
        if atom.GetAtomMapNum() != 0:
            return True
    return False


def smiles_has_atom_mapping(smiles: str) -> bool:
    """
    Whether at least one atom of a compound SMILES contains an atom map number.

    Args:
        smiles: compound SMILES.
    """
    mol = smiles_to_mol(smiles, sanitize=False)
    return mol_has_atom_mapping(mol)



import logging
from pathlib import Path
from typing import Iterable, Iterator

import click
from rxn.utilities.files import dump_list_to_file
from rxn.utilities.logging import setup_console_logger

from ..rdf import RdfReaction, ReactionSmilesExtractor, iterate_reactions_from_file

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@click.command()
@click.option(
    "-r",
    "--rdf_file",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=True,
    help="RDF file (to read from).",
)
@click.option(
    "-s",
    "--smiles_file",
    type=click.Path(writable=True, path_type=Path),
    required=True,
    help="Where to save the reactions (in SMILES format, one per line).",
)
@click.option(
    "--fragment_bond",
    type=str,
    default="~",
    help="Fragment bond.",
)
def main(rdf_file: Path, smiles_file: Path, fragment_bond: str) -> None:
    """Convert a file of RDF reactions to SMILES format."""
    setup_console_logger()

    extractor = ReactionSmilesExtractor(fragment_bond=fragment_bond, sanitize=False)
    total_reactions = 0
    successful_reactions = 0

    def convert(rdfs: Iterable[RdfReaction]) -> Iterator[str]:
        """Convert reactions, ignoring the ones causing an error.

        Note: should be refactored to a class if the functionality is
        to be made a bit cleaner (avoid ``nonlocal``)."""
        nonlocal total_reactions
        nonlocal successful_reactions

        for rdf in rdfs:
            total_reactions += 1
            try:
                yield extractor.to_reaction_smiles(rdf)
                successful_reactions += 1
            except Exception as e:
                logger.warning(f"Cannot convert reaction: {e}")
                continue

    rdf_reactions = iterate_reactions_from_file(rdf_file)
    smiles_reactions = convert(rdf_reactions)
    dump_list_to_file(smiles_reactions, smiles_file)

    logger.info(
        f"Finished conversion. Successful: {successful_reactions} / {total_reactions}."
    )


if __name__ == "__main__":
    main()



import pytest
from rdkit import Chem
from rdkit.Chem import Mol

from rxn.chemutils.conversion import (
    canonicalize_smiles,
    cleanup_smiles,
    inchi_to_mol,
    maybe_canonicalize,
    mdl_to_mol,
    mdl_to_smiles,
    mol_to_inchi,
    mol_to_mdl,
    mol_to_smiles,
    remove_hydrogens,
    sanitize_mol,
    smiles_to_inchi,
    smiles_to_mol,
    split_smiles_and_fragment_info,
)
from rxn.chemutils.exceptions import (
    InvalidInchi,
    InvalidMdl,
    InvalidSmiles,
    SanitizationError,
)

WATER_MDL = """
     RDKit          2D

  1  0  0  0  0  0  0  0  0  0999 V2000
    0.0000    0.0000    0.0000 O   0  0  0  0  0  0  0  0  0  0  0  0
M  END
"""

CFC_MDL = """
     RDKit          2D

  3  2  0  0  0  0  0  0  0  0999 V2000
    0.0000    0.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.2990    0.7500    0.0000 F   0  0  0  0  0  0  0  0  0  0  0  0
    2.5981   -0.0000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0
  2  3  1  0
M  END
"""


def test_smiles_to_mol() -> None:
    mol = smiles_to_mol("COC")
    assert mol.GetNumAtoms() == 3
    assert mol.GetAtomWithIdx(0).GetSymbol() == "C"
    assert mol.GetAtomWithIdx(1).GetSymbol() == "O"
    assert mol.GetAtomWithIdx(2).GetSymbol() == "C"

    # must raise for invalid SMILES (RDKit's default: return None)
    with pytest.raises(InvalidSmiles):
        smiles_to_mol("")

    # Invalid SMILES symbol
    with pytest.raises(InvalidSmiles):
        smiles_to_mol("L")

    # Invalid valence
    with pytest.raises(InvalidSmiles):
        smiles_to_mol("CFC")

    # Radicals are kept when converting back to a SMILES
    assert mol_to_smiles(smiles_to_mol("CC[C]CC")) == "CC[C]CC"


def test_smiles_to_mol_without_sanitization() -> None:
    # With sanitization, this would fail
    mol = smiles_to_mol("CFC", sanitize=False)
    assert mol.GetNumAtoms() == 3
    assert mol.GetAtomWithIdx(0).GetSymbol() == "C"
    assert mol.GetAtomWithIdx(1).GetSymbol() == "F"
    assert mol.GetAtomWithIdx(2).GetSymbol() == "C"

    # empty molecule
    with pytest.raises(InvalidSmiles):
        smiles_to_mol("", sanitize=False)

    # invalid SMILES symbol
    with pytest.raises(InvalidSmiles):
        smiles_to_mol("L", sanitize=False)

    # Doing strictly no sanitization with the RDKit functions messes up the
    # radicals - smiles_to_mol takes care of this
    assert mol_to_smiles(smiles_to_mol("CC[C]CC", sanitize=False)) == "CC[C]CC"


def test_mol_to_smiles_with_isomeric_smiles() -> None:
    mol = smiles_to_mol("N[C@@](F)(C)C(=O)O")

    # Test without isomeric smiles
    assert mol_to_smiles(mol, isomericSmiles=False) == "CC(N)(F)C(=O)O"

    # Test with isomeric smiles
    assert mol_to_smiles(mol, isomericSmiles=True) == "C[C@@](N)(F)C(=O)O"


def test_inchi_to_mol() -> None:
    mol = inchi_to_mol("InChI=1S/C2H6O/c1-3-2/h1-2H3")
    assert mol.GetNumAtoms() == 3
    assert mol.GetAtomWithIdx(0).GetSymbol() == "C"
    assert mol.GetAtomWithIdx(1).GetSymbol() == "C"
    assert mol.GetAtomWithIdx(2).GetSymbol() == "O"

    # must raise for invalid InChI (RDKit's default: return None)
    with pytest.raises(InvalidInchi):
        inchi_to_mol("")

    # Invalid InChI symbol
    with pytest.raises(InvalidInchi):
        inchi_to_mol("L")

    # Test conversion to and from mol
    assert (
        mol_to_inchi(inchi_to_mol("InChI=1S/C2H6O/c1-3-2/h1-2H3"))
        == "InChI=1S/C2H6O/c1-3-2/h1-2H3"
    )


def test_mdl_to_mol() -> None:
    # NB: we assert that the molecules are correct with their SMILES notation
    assert mol_to_smiles(mdl_to_mol(WATER_MDL)) == "O"

    # Empty MDL
    with pytest.raises(InvalidMdl):
        _ = mdl_to_mol("")

    # Invalid valence - raises by default
    with pytest.raises(InvalidMdl):
        _ = mdl_to_mol(CFC_MDL)

    # Invalid valence - does not raise if no sanitization
    assert mol_to_smiles(mdl_to_mol(CFC_MDL, sanitize=False)) == "CFC"

    # Back-conversion leads to original MDL
    assert mol_to_mdl(mdl_to_mol(WATER_MDL)) == WATER_MDL
    assert mol_to_mdl(mdl_to_mol(CFC_MDL, sanitize=False)) == CFC_MDL


def test_mdl_to_smiles() -> None:
    assert mdl_to_smiles(WATER_MDL) == "O"

    # invalid MDL -> exception
    with pytest.raises(InvalidMdl):
        _ = mdl_to_smiles("invalid mdl")

    # invalid MDL, called with default -> returns the default
    assert mdl_to_smiles("invalid mdl", default="[Invalid]") == "[Invalid]"


def test_sanitize_mol() -> None:
    def example_mol() -> Mol:
        return smiles_to_mol("C1=CC=CC=C1N(=O)=O", sanitize=False)

    # Example 1: do only aromatization
    mol = example_mol()
    sanitize_mol(mol, include_sanitizations=[Chem.SANITIZE_SETAROMATICITY])
    assert mol_to_smiles(mol, canonical=False) == "c1ccccc1N(=O)=O"

    # Example 2: do only cleanup
    mol = example_mol()
    sanitize_mol(mol, include_sanitizations=[Chem.SANITIZE_CLEANUP])
    assert mol_to_smiles(mol, canonical=False) == "C1=CC=CC=C1[N+]([O-])=O"

    # Example 3: do both cleanup and aromatization
    mol = example_mol()
    sanitize_mol(
        mol, include_sanitizations=[Chem.SANITIZE_CLEANUP, Chem.SANITIZE_SETAROMATICITY]
    )
    assert mol_to_smiles(mol, canonical=False) == "c1ccccc1[N+]([O-])=O"

    # Example 6: do all except aromatization
    mol = example_mol()
    sanitize_mol(mol, exclude_sanitizations=[Chem.SANITIZE_SETAROMATICITY])
    assert mol_to_smiles(mol, canonical=False) == "C1=CC=CC=C1[N+]([O-])=O"

    # Example 5: do all except cleanup (and properties -> invalid valence otherwise)
    mol = example_mol()
    sanitize_mol(
        mol, exclude_sanitizations=[Chem.SANITIZE_CLEANUP, Chem.SANITIZE_PROPERTIES]
    )
    assert mol_to_smiles(mol, canonical=False) == "c1ccccc1N(=O)=O"

    # Raises exception for invalid sanitization
    mol = smiles_to_mol("CFC", sanitize=False)
    with pytest.raises(SanitizationError):
        sanitize_mol(mol, include_sanitizations=[Chem.SANITIZE_PROPERTIES])

    # Not providing any options does the full sanitization
    mol_1 = example_mol()
    mol_2 = example_mol()
    sanitize_mol(mol_1)
    sanitize_mol(mol_2, include_sanitizations=[Chem.SANITIZE_ALL])
    assert mol_to_smiles(mol_1) == mol_to_smiles(mol_2)

    # Cannot specify all included and excluded sanitizations at the same time
    with pytest.raises(ValueError):
        sanitize_mol(
            example_mol(),
            include_sanitizations=[Chem.SANITIZE_ALL],
            exclude_sanitizations=[Chem.SANITIZE_NONE],
        )


def test_remove_hydrogens() -> None:
    smiles = "[H]C([H])([H])Oc1ccc2cccc(C([H])([H])C#N)c2c1"
    smiles_without_h = "COc1ccc2cccc(CC#N)c2c1"

    # Removing the unnecessary hydrogens
    mol = smiles_to_mol(smiles, sanitize=False)
    assert mol_to_smiles(mol, canonical=False) == smiles
    mol = remove_hydrogens(mol)
    assert mol_to_smiles(mol, canonical=False) == smiles_without_h

    # check that it does not sanitize the molecules
    smiles = "[H]C([H])([H])CN(=O)=O"
    smiles_without_h = "CCN(=O)=O"
    mol = smiles_to_mol(smiles, sanitize=False)
    mol = remove_hydrogens(mol)
    assert mol_to_smiles(mol, canonical=False) == smiles_without_h


def test_canonicalize() -> None:
    assert canonicalize_smiles("CCO") == canonicalize_smiles("OCC")

    # An empty SMILES string must raise an error
    with pytest.raises(InvalidSmiles):
        canonicalize_smiles("")

    # invalid SMILES symbol
    with pytest.raises(InvalidSmiles):
        canonicalize_smiles("L")

    # invalid valence
    with pytest.raises(InvalidSmiles):
        canonicalize_smiles("CFC")

    # Canonicalization is possible for invalid valence if one does not sanitize
    assert canonicalize_smiles("CFC", check_valence=False) == "CFC"
    assert canonicalize_smiles("F(C)C", check_valence=False) == "CFC"
    assert canonicalize_smiles("F(C)(C)", check_valence=False) == "CFC"

    # For invalid valences, still canonicalizes aromaticity, etc.
    assert canonicalize_smiles(
        "C1=CC=CC=C1CFC", check_valence=False
    ) == canonicalize_smiles("c1ccccc1CFC", check_valence=False)
    # For special aromatic cases, still canonicalizes aromaticity, etc.
    assert canonicalize_smiles(
        "[se]1cccc1", check_valence=False
    ) == canonicalize_smiles("[se]1C=CC=C1", check_valence=False)


def test_canonicalize_rotation() -> None:
    # In an earlier version, the following SMILES kept switching between two
    # SMILES variants when canonicalizing (back-and-forth between aromatic and
    # non-aromatic.
    smiles = "c1cc[se]c1"
    canonical_1 = canonicalize_smiles(smiles)
    canonical_2 = canonicalize_smiles(canonical_1)
    assert canonical_2 == canonical_1


def test_canonicalize_removes_unnecessary_hydrogens() -> None:
    # in an earlier version, canonicalize_smiles(x) did not behave the same as
    # Chem.MolToSmiles(Chem.MolFromSmiles(x)) when there were implicit hydrogens.
    smiles = "[H]C([H])([H])Oc1ccc2cccc(C([H])([H])C#N)c2c1"

    assert canonicalize_smiles(smiles) == "COc1ccc2cccc(CC#N)c2c1"


def test_maybe_canonicalize() -> None:
    # basic example
    assert maybe_canonicalize("C(C)C") == "CCC"

    # invalid -> does not change the SMILES
    assert maybe_canonicalize("C(CC") == "C(CC"
    assert maybe_canonicalize("F(C)C") == "F(C)C"

    # invalid valence, but flag given -> does the canonicalization
    assert maybe_canonicalize("F(C)C", check_valence=False) == "CFC"


def test_cleanup_smiles() -> None:
    # Basic checks
    assert cleanup_smiles("C") == "C"
    assert cleanup_smiles("[C]") == "[C]"
    assert cleanup_smiles("[CH4]") == "C"
    assert cleanup_smiles("C[C]C") == "C[C]C"
    assert cleanup_smiles("C[CH2]C") == "CCC"

    # Removes unnecessary dashes and parentheses
    assert cleanup_smiles("C-C") == "CC"
    assert cleanup_smiles("C(C)") == "CC"

    # Does not do any canonicalization
    assert cleanup_smiles("C(C)O") == "C(C)O"

    # Does not do any kekulization / aromatization
    assert cleanup_smiles("c1ccccc1") == "c1ccccc1"
    assert cleanup_smiles("C1=CC=CC=C1") == "C1=CC=CC=C1"
    assert cleanup_smiles("[se]1cccc1") == "[se]1cccc1"
    assert cleanup_smiles("[se]1C=CC=C1") == "[se]1C=CC=C1"

    # Does not mess up with fragment molecules
    assert cleanup_smiles("[C].[Pd]") == "[C].[Pd]"
    assert cleanup_smiles("[Pd].[C]") == "[Pd].[C]"

    # Does no valence check
    carbon_with_eight_neighbours = "C(C)(C)(C)(C)(C)(C)(C)C"
    assert cleanup_smiles(carbon_with_eight_neighbours) == carbon_with_eight_neighbours
    na_with_two_neighbors = "[Na]1OS(=O)(=O)O1"
    assert cleanup_smiles(na_with_two_neighbors) == na_with_two_neighbors

    # More complex example. It is not possible to avoid some reordering here,
    # it should actually be CC(C)c1ccc(cc1)C(=O)CCCCl.
    assert (
        cleanup_smiles(
            "[CH3][CH]([CH3])[c]1[cH][cH][c]([cH][cH]1)[C](=[O])[CH2][CH2][CH2][Cl]"
        )
        == "CC(C)c1ccc(C(=O)CCCCl)cc1"
    )

    # Raises for invalid atom types
    with pytest.raises(InvalidSmiles):
        cleanup_smiles("CC[Invalid]O")

    # Raises for empty string
    with pytest.raises(InvalidSmiles):
        cleanup_smiles("")


def test_split_reaction_smiles() -> None:
    reaction_smiles = (
        "BrC1=C(C(=O)O)C=C(C=C1)COC.C1(O)=CC(O)=CC=C1.[OH-].[Na+]"
        ">S(=O)(=O)([O-])[O-].[Cu+2].O"
        ">OC1=CC=C2C3=C(C(OC2=C1)=O)C=C(C=C3)COC"
        " |f:2.3,4.5|"
    )

    pure_smiles, fragment_info = split_smiles_and_fragment_info(reaction_smiles)
    assert (
        pure_smiles == "BrC1=C(C(=O)O)C=C(C=C1)COC.C1(O)=CC(O)=CC=C1.[OH-].[Na+]"
        ">S(=O)(=O)([O-])[O-].[Cu+2].O"
        ">OC1=CC=C2C3=C(C(OC2=C1)=O)C=C(C=C3)COC"
    )
    assert fragment_info == "|f:2.3,4.5|"


def test_split_reaction_smiles_with_no_fragment_info() -> None:
    reaction_smiles = (
        "[CH3:7][C:4]1=[CH:3][C:2](=[C:1]([CH:6]=[CH:5]1)Br)[F:8]"
        ">[Li]CCCC.C1CCOC1"
        ">[CH3:7][C:4]1=[CH:3][C:2](=[C:1]([CH:6]=[CH:5]1)C(=O)O)[F:8]"
    )

    pure_smiles, fragment_info = split_smiles_and_fragment_info(reaction_smiles)
    assert pure_smiles == reaction_smiles
    assert fragment_info == ""


def test_smiles_to_inchi() -> None:
    # different SMILES representations for phenol (aromatic, non-aromatic)
    # should be converted to the same InChI
    phenol_inchi = "InChI=1S/C6H6O/c7-6-4-2-1-3-5-6/h1-5,7H"
    for smiles in ["c1ccccc1O", "C1=CC=CC=C1O"]:
        assert smiles_to_inchi(smiles) == phenol_inchi

    # different SMILES representations for 4(1H)-pyrimidinone (tautomers)
    # should be converted to the same InChI
    expected_inchi = "InChI=1S/C4H4N2O/c7-4-1-2-5-3-6-4/h1-3H,(H,5,6,7)"
    for smiles in ["C1=CN=CNC1=O", "C1=CNC=NC1=O", "C1=CN=CN=C1O"]:
        assert smiles_to_inchi(smiles) == expected_inchi

    with pytest.raises(InvalidSmiles):
        smiles_to_inchi("")

    with pytest.raises(InvalidSmiles):
        smiles_to_inchi("invalid")

    # Does not raise for invalid valence
    assert smiles_to_inchi("CFC") == "InChI=1S/C2H6F/c1-3-2/h1-2H3"

    # SMILES that are aromatized or not lead to the same InChi
    assert smiles_to_inchi("C1=CC=CC=C1") == smiles_to_inchi("c1ccccc1")

    # Stereochemistry information is kept: the InChi should be different if the
    # SMILES have different double-bond or R/S stereochemistries
    assert len({smiles_to_inchi(s) for s in ["CC=CC", "C/C=C/C", r"C/C=C\C"]}) == 3
    assert (
        len({smiles_to_inchi(s) for s in ["C(O)(N)C", "[C@H](O)(N)C", "[C@@H](O)(N)C"]})
        == 3
    )

    # extended tautomer check - where the default options do not detect
    # tautomerism in a 1,3-diketone
    assert smiles_to_inchi("C1C(=O)CC(=O)CC1") != smiles_to_inchi("C1C(=O)C=C(O)CC1")
    assert smiles_to_inchi("C1C(=O)CC(=O)CC1", True) == smiles_to_inchi(
        "C1C(=O)C=C(O)CC1", True
    )



from typing import Dict, List

from attr import define


@define
class RdfReaction:
    """
    Custom class wrapping an RDF reaction.

    Attributes:
        reactants: List of reactants, given in MDL format.
        reagents: List of reagents, given in MDL format.
        products: List of products, given in MDL format.
        meta: meta information.
        reaction_index: index of the reaction, coming from the "$RFMT $RIREG" line.
    """

    reactants: List[str]
    reagents: List[str]
    products: List[str]
    meta: Dict[str, str]
    reaction_index: int

    def get_mol_structure(self, category: str, index: int) -> str:
        """
        Get a structure given in the reaction metadata, such as in
        RXN:CATALYST(2):MOL(1):MOLSTRUCTURE

        Args:
            category: 'catalyst' for the example above.
            index: 2 for the example above.

        Returns:
            The MolBlock for the desired structure.
        """
        key = f"RXN:{category.upper()}({index}):MOL(1):MOLSTRUCTURE"
        return self.meta[key]

    def additional_reagents_from_meta(self) -> List[str]:
        """
        Get the MDL strings for the additional reagents from the meta information.
        """
        return [
            value for key, value in self.meta.items() if key.endswith("MOLSTRUCTURE")
        ]

    def all_reagents(self) -> List[str]:
        """
        Get the MDL strings for all the reagents.
        """

        reagent_mdls = self.reagents.copy()
        reagent_mdls.extend(self.additional_reagents_from_meta())
        return reagent_mdls



import random
from typing import Callable, List

from rxn.utilities.containers import all_identical

from rxn.chemutils.conversion import canonicalize_smiles
from rxn.chemutils.smiles_randomization import (
    randomize_smiles_restricted,
    randomize_smiles_rotated,
    randomize_smiles_unrestricted,
)

randomization_functions: List[Callable[[str], str]] = [
    randomize_smiles_unrestricted,
    randomize_smiles_restricted,
    randomize_smiles_rotated,
]


def test_generates_all_randomizations_for_simple_example() -> None:
    # Check that everything is recovered in a simple example with all methods
    smiles = "C1ON1"

    # 6 possible SMILES: 3 different starting atoms x 2 directions
    expected = {"C1ON1", "C1NO1", "O1NC1", "O1CN1", "N1CO1", "N1OC1"}
    for fn in randomization_functions:
        samples = {fn(smiles) for _ in range(100)}
        assert samples == expected

    # randomizing with rotation and no order reversal: only 3 possibilities
    assert (
        len(
            {
                randomize_smiles_rotated(smiles, with_order_reversal=False)
                for _ in range(100)
            }
        )
        == 3
    )


def test_does_not_sanitize_molecules() -> None:
    # Check that no aromatization or cleanup is done
    smiles = "C1=CC=CC=C1N(=O)=O"

    for fn in randomization_functions:
        samples = {fn(smiles) for _ in range(10)}
        assert not any("c" in sample for sample in samples)
        assert not any("[N+]" in sample for sample in samples)

    # Check that radicals are kept
    smiles = "CC[C]CC"
    for fn in randomization_functions:
        samples = {fn(smiles) for _ in range(10)}
        assert all("[C]" in sample for sample in samples)


def test_does_not_change_aromaticity() -> None:
    # Check that there is no conversion of aromatic vs kekulized representation
    # (Selenium caused a few problems in other contexts)
    smiles = "c1cc[se]c1"
    for fn in randomization_functions:
        samples = {fn(smiles) for _ in range(10)}
        assert all("[se]" in sample for sample in samples)
    smiles = "[Se]1C=CC=C1"
    for fn in randomization_functions:
        samples = {fn(smiles) for _ in range(10)}
        assert all("[Se]" in sample for sample in samples)


def test_keeps_stereochemistry() -> None:
    # Simple molecule with a stereocenter
    smiles = "COC[C@](CNC)(Cl)Br"

    for fn in randomization_functions:
        samples = {fn(smiles) for _ in range(10)}
        # All the samples must have kept at least one '@'
        assert all("@" in sample for sample in samples)


def test_number_generated_molecules() -> None:
    # For a test molecule, verify that the unrestricted randomization generates
    # most molecules, followed by the restricted one, followed by the rotated one.
    # NB: the exact values below depend on the number of samples.

    adenine = "Nc1ncnc2[nH]cnc12"
    n_samples = 300

    rotated = {randomize_smiles_rotated(adenine) for _ in range(n_samples)}
    restricted = {randomize_smiles_restricted(adenine) for _ in range(n_samples)}
    unrestricted = {randomize_smiles_unrestricted(adenine) for _ in range(n_samples)}

    assert 18 < len(rotated) < 22
    assert 45 < len(restricted) < 60
    assert 90 < len(unrestricted)


def test_reproducibility() -> None:
    adenine = "Nc1ncnc2[nH]cnc12"

    for fn in randomization_functions:
        samples = []
        for _ in range(10):
            random.seed(42)
            samples.append(fn(adenine))

        assert all_identical(samples)

        # sampling one more time without resetting the seed -> different
        samples.append(fn(adenine))
        assert not all_identical(samples)


def test_multi_fragment_compounds() -> None:
    # Check that no aromatization or cleanup is done
    smiles = "CC[NH+].CC(=O)[O-]"

    for fn in randomization_functions:
        samples = {fn(smiles) for _ in range(20)}

        # Make sure that some randomization is happening
        assert len(samples) > 5

        # make sure that the order of the fragments is shuffled; check that
        # N is present sometimes in the first fragment, sometimes in the second.
        assert any("N" in sample.split(".")[0] for sample in samples)
        assert any("N" in sample.split(".")[1] for sample in samples)

        # All of them should still have the same canonical representation
        assert len({canonicalize_smiles(sample) for sample in samples}) == 1



from rxn.chemutils.conversion import mol_to_smiles, smiles_to_mol
from rxn.chemutils.rdkit_utils import combine_mols


def test_combine_mols() -> None:
    # No molecule -> empty Mol
    assert mol_to_smiles(combine_mols([])) == ""

    # Test a few single mols
    for smiles in ["CC", "CCO", "CC(C)CCO"]:
        mol = smiles_to_mol(smiles)
        assert mol_to_smiles(combine_mols([mol])) == smiles

    # Test a few combinations of molecules
    for multi_smiles in [
        ("CC", "O"),
        ("CC", "CCO", "N"),
        ("CC", "CC", "CC(C)CCO", "c1ccccc1"),
    ]:
        mols = (smiles_to_mol(smiles) for smiles in multi_smiles)
        merged_mol = combine_mols(mols)
        # we verify the merge by checking that the conversion back to SMILES
        # leads to the '.'-concatenated separate SMILES.
        assert mol_to_smiles(merged_mol) == ".".join(multi_smiles)
